from dungeonsheets.spells.spells import Spell


class Bane(Spell):
    """Up to three creatures of your choice that you can see within range must make
    Charisma saving throws. Whenever a target that fails this saving throw makes an
    attack roll or a saving throw before the spell ends, the target must roll a d4
    and subtract the number rolled from the attack roll or saving throw.

    At Higher
    Levels: When you cast this spell using a spelslot of 2nd level or higher, you
    can target one aditional creature for each slot level above 1st.
    """

    name = "Bane"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A drop of blood"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Cleric")


class BanishingSmite(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends,
    your weapon crackles with force, and the attack deals an extra 5d10 force damage
    to the target. Additionally, if this attack reduces the target to 50 hit points
    of fewer, you banish it. If the target is native to a different plane of
    existence than the on you're on, the target disappears, returning to its home
    plane. If the target is native to the plane you're on, the creature vanishes
    into a harmless demiplane. While there, the target is incapacitated. It remains
    there until the spell ends, at which point the target reappears in the space it
    left or in the nearest unoccupied space if that space is occupied.
    """

    name = "Banishing Smite"
    level = 5
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Paladin",)


class Banishment(Spell):
    """You attempt to send one creature that you can see within range to another place
    of existence. The target must succeed on a Charisma saving throw or be banished.


    If the target is native to the plane of existence you're on, you banish the
    target to a harmless demiplane. While there, the target is incapacitated. The
    target remains there until the spell ends, at which point the target reappears
    in the space it left or in the nearest unoccupied space if that space is
    occupied.

    If the target is native to a different plane of existence that the
    one you're on, the target is banished with a faint popping noise, returning to
    its home plane.
    If the spell ends before 1 minute has passed, the target
    reappears in the space it left or in the nearest unoccupied space if that space
    is occupied. Otherwise, the target doesn't return.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 5th level or higher, you can target one
    additional creature for each slot level above 4th.
    """

    name = "Banishment"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "An item distasteful to the target"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin", "Sorcerer", "Warlock", "Wizard")


class Barkskin(Spell):
    """You touch a willing creature. Until the spellends, the target's skin has a
    rough, bark-like appearance, and the target's AC can't be less than 16,
    regardless of what kind of armor it is wearing.
    """

    name = "Barkskin"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A handful of oak bark"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Ranger")


class BeaconOfHope(Spell):
    """This spell bestows hope and vitality. Choose any number of creatures within
    range. For the duration, each target has advantage on Wisdom saving throws and
    death saving throws, and regains the maximum number of hit points possible from
    any healing.
    """

    name = "Beacon Of Hope"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric",)


class BeastBond(Spell):
    """You establish a telepathic link with one beast you touch that is
    friendly to you or charmed by you. The spell fails if the beast's
    Intelligence is 4 or higher.  Until the spell ends, the link is
    active while you and the beast are within line of sight of each
    other. Through the link, the beast can understand your telepathic
    messages to it, and it can telepathically communicate simple
    emotions and concepts back to you. While the link is active, the
    beast gains advantage on attack rolls against any creature within
    5 feet of you that you can see.

    """

    name = "Beast Bond"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A bit of fur wrapped in a cloth"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Divination"
    classes = ("Druid", "Ranger")


class BeastSense(Spell):
    """You touch a willing beast. For the duration of the spell, you can
    use your action to see through the beast's eyes and hear what it
    hears, and continue to do so until you use your action to return
    to your normal senses.

    """

    name = "Beast Sense"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("S",)
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = True
    magic_school = "Divination"
    classes = ("Druid", "Ranger")


class BestowCurse(Spell):
    """You touch a creature, and that creature must succeed on a Wisdom saving throw or
    become cursed for the duration of the spell. When you cast this spell, choose
    the nature of the curse from the following options:* Choose one ability score.
    While cursed, the target has disadvantage on ability checks and saving throws
    made with that ability score.* While cursed, the target has disadvantage on
    attack rolls against you.* While cursed, the target must make a Wisdom saving
    throw at the start of each of its turns. If it fails, it wastes its action that
    turn doing nothing.* While the target is cursed, your attacks and spells deal an
    extra 1d8 necrotic damage to the target.

    A remove curse spell ends this
    effect. At the DM's option, you may choose an alternative curse effect, but it
    should be no more powerful than those described above.
    The DM has final say on
    such a curse's effect.

    **At Higher Levels:** If you cast this spell using a spell
    slot of 4th level or higher, the duration is concentration, up to 10 minutes.

    If you use a spell slot of 5th level or higher, the duration is 8 hours.
    If
    you use a spell slot of 7th level or higher, the duration is 24 hours.
    If you
    use a 9th level spell slot, the spell lasts until it is dispelled.
    Using a
    spell slot of 5th level or higher grants a duration that doesn't require
    concentration.
    """

    name = "Bestow Curse"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Bard", "Cleric", "Wizard")


class BigbysHand(Spell):
    """You create a Large hand of shimmering, translucent force in an
    unoccupied space that you can see within range. The hand lasts for
    the spell's duration, and it moves at your command, mimicking the
    movements of your own hand.

    The hand is an object that has AC 20 and hit points equal to your
    hit point maximum. If it drops to 0 hit points, the spell ends. It
    has a Strength of 26 (+8) and a Dexterity of 10 (+0). The hand
    doesn't fill its space.

    When you cast the spell and as a bonus action on your subsequent
    turns, you can move the hand up to 60 feet and then cause one of
    the following effects with it.

    **Clenched Fist**
    The hand strikes one creature or object within 5 feet of it. Make
    a melee spell attack for the hand using your game statistics. On a
    hit, the target takes 4d8 force damage.

    **Forceful Hand**
    The hand attempts to push a creature within 5 feet of it in a
    direction you choose. Make a check with the hand's Strength
    contested by the Strength (Athletics) check of the target. If the
    target is Medium or smaller, you have advantage on the check. If
    you succeed, the hand pushes the target up to 5 feet plus a number
    of feet equal to five times your spellcasting ability
    modifier. The hand moves with the target to remain within 5 feet
    of it.

    **Grasping Hand**
    The hand attempts to grapple a Huge or smaller creature within 5
    feet of it. You use the hand's Strength score to resolve the
    grapple. If the target is Medium or smaller, you have advantage on
    the check.  While the hand is grappling the target, you can use a
    bonus action to have the hand crush it. When you do so, the target
    takes bludgeoning damage equal to 2d6 + your spellcasting ability
    modifier.

    **Interposing Hand**
    The hand interposes itself between you and a creature you choose
    until you give the hand a different command. The hand moves to
    stay between you and the target, providing you with half cover
    against the target. The target can't move through the hand's space
    if its Strength score is less than or equal to the hand's Strength
    score. If its Strength score is higher than the hand's Strength
    score, the target can move toward you through the hand's space,
    but that space is difficult terrain for the target.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 6th level or higher, the damage from the clenched fist option
    increases by 2d8 and the damage from the grasping hand increases
    by 2d6 for each slot level above 5th.

    """

    name = "Bigbys Hand"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S", "M")
    materials = "An eggshell and a snakeskin glove"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Wizard",)


class BladeBarrier(Spell):
    """You create a vertical wall of whirling, razor-sharp blades made of magical
    energy. The wall appears within range and lasts for the duration. You can make a
    straight wall up to 100 feet long, 20 feet high, and 5 feet thick, or a ringed
    wall up to 60 feet in diameter, 20 feet high, and 5 feet thick. The wall
    provides three-quarters cover to creatures behind it, and its space is difficult
    terrain.

     When a creature enters the wall's area for the first time on a turn
    or starts its turn there, the creature must make a Dexterity saving throw. On a
    failed save, the creature takes 6 d10 slashing damage. On a successful save,
    the creature takes half as much damage.
    """

    name = "Blade Barrier"
    level = 6
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric",)


class BladeWard(Spell):
    """You extend your hand and trace a sigil of warding in the air. Until the end of
    your next turn, you have resistance against bludgeoning, piercing, and slashing
    damage dealt by weapon attacks.
    """

    name = "Blade Ward"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "1 round"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class Bless(Spell):
    """You bless up to three creatures of your choice within range. Whenever a target
    makes an attack roll or a saving throw before the spell ends, the target can
    roll a d4 and add the number rolled to the attack roll or saving throw.

    At
    Higher Levels: When you cast this spell using a spell slot of 2nd level or
    higher, you can target one additional creature for each slot level above 1st.
    """

    name = "Bless"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A sprinkling of holy water"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Cleric", "Paladin")


class Blight(Spell):
    """Necromantic energy washes over a creature of your choice that you can see within
    range, draining moisture and vitality from it. The target must make a
    Constitution saving throw. The target takes 8d8 necrotic damage on a failed
    save, or half as much damage on a successful one. This spell has no effect on
    undead or constructs.

    If you target a plant creature or a magical plant, it
    makes the saving throw with disadvantage, and the spell deals maximum damage to
    it.

    If you target a nonmagical plant that isn't a creature, such as a tree or
    shrub, it doesn't make a saving throw; it simply withers and dies.

    At Higher
    Levels: When you cast this spell using a spell slot of 5th level or higher, the
    damage increases by 1d8 for each slot level above 4th.
    """

    name = "Blight"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class BlindingSmite(Spell):
    """The next time you hit a creature with a melee weapon attack during
    this spell's duration, you weapon flares with a bright light, and
    the attack deals an extra 3d8 radiant damage to the
    target. Additionally, the target must succeed on a Constitution
    saving throw or be blinded until the spell ends.

    A creature blinded by this spell makes another Constitution saving
    throw at the end of each of its turns. On a successful save, it is
    no longer blinded.

    """

    name = "Blinding Smite"
    level = 3
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Paladin",)


class BlindnessDeafness(Spell):
    """You can blind or deafen a foe. Choose one creature that you can see
    within range to make a Constitution saving throw. If it fails, the
    target is either blinded or deafened (your choice) for the
    duration. At the end of each of its turns, the target can make a
    Constitution saving throw. On a success, the spell ends.

    **At Higher Levels:** When you cast this spell using a spell slot of
    3rd level or higher, you can target one additional creature for
    each slot level above 2nd.

    """

    name = "Blindness/Deafness"
    level = 2
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V",)
    duration = "1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Bard", "Cleric", "Sorcerer", "Wizard")


class Blink(Spell):
    """Roll a d20 at the end of each of your turns for the duration of the spell. On a
    roll of 11 or higher, you vanish from your current plane of existence and appear
    in the Etheral Plane (the spell fails and the casting is wasted if you were
    already on that plane).

    At the start of you next turn, and when the spell ends
    if you are on the Etheral Plane, you return to an unoccupied space of your
    choice that you can see within 10 feet of the space you vanished from. If no
    unoccupied space is available within that rang, you appear in the nearest
    unoccupied space (chosen at random if more that one space is equally near). You
    can dismiss this spell as an action.

    While on the Ethereal Plane, you can see
    and hear the plane you originated from, which is cast in shades of gray, and you
    can't see anything more than 60 feet away.You can only affect and be affected
    by other reatures on the Ethereal Plane. Creature that aren't there can't
    perceive you or interact with you, unless they have the ability to do so.
    """

    name = "Blink"
    level = 3
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class Blur(Spell):
    """Your body becomes blurred, shifting and wavering to all who can see you. For the
    duration, any creature has disadvantage on attack rolls against you. An
    attacker is immune to this effect if it doesnt rely on sight, as with
    blindsight, or can see through illusions, as with truesight.
    """

    name = "Blur"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ("Sorcerer", "Wizard")


class BonesOfTheEarth(Spell):
    """You cause up to six pillars of stone to burst from places on the
    ground that you can see within range. Each pillar is a cylinder
    that has a diameter of 5 feet and a height of up to 30 feet. The
    ground where a pillar appears must be wide enough for its
    diameter, and you can target ground under a creature if that
    creature is Medium or smaller. Each pillar has AC 5 and 30 hit
    points. When reduced to 0 hit points, a pillar crumbles into
    rubble, which creates an area of difficult terrain with a 10-foot
    radius. The rubble lasts until cleared.

    If a pillar is created under a creature, that creature must
    succeed on a Dexterity saving throw or be lifted by the pillar. A
    creature can choose to fail the save.

    If a pillar is prevented from reaching its full height because of
    a ceiling or other obstacle, a creature on the pillar takes 6d6
    bludgeoning damage and is restrained, pinched between the pillar
    and the obstacle. The restrained creature can use an action to
    make a Strength or Dexterity check (the creature's choice) against
    the spell's saving throw DC. On a success, the creature is no
    longer restrained and must either move off the pillar or fall off
    it.

    **At Higher Levels.** When you cast this spell using a spell slot
    of 7th level or higher, you can create two additional pillars for
    each slot level above 6th.

    """

    name = "Bones Of The Earth"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid",)


class BoomingBlade(Spell):
    """As part of the action used to cast this spell, you must make a melee attack with
    a weapon against one creature within the spell's range, otherwise the spell
    fails.
    On a hit, the target suffers the attack's normal effects, and it becomes
    sheathed in booming energy until the start of your next turn. If the target
    willingly moves be- fore then, it immediately takes 1d8 thunder damage, and the
    spell ends.
    This spell's damage increases when you reach higher levels.

    At
    Higher Levels: At 5th level, the melee attack deals an extra 1d8 thunder damage
    to the target, and the damage the target takes for moving increases to 2d8. Both
    damage rolls increase by 1d8 at 11th level and 17th level.
    """

    name = "Booming Blade"
    level = 0
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ("V", "M")
    materials = "A weapon"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Warlock", "Wizard")


class BrandingSmite(Spell):
    """The next time you hit a creature with a weapon attack before this spell ends,
    the weapon glemas with astral radiance as you strike. The attack deals an extra
    2d6 radiant damage to the target, which becomes visible if it is invisible, and
    the target sheds dim light in a 5-foot radius and can't become invisible until
    the spell ends.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 3rd level or higher, the extra damage increases by 1d6 for each slot level
    above 2nd.
    """

    name = "Branding Smite"
    level = 2
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Paladin",)


class BurningHands(Spell):
    """As you hold your hands with thumbs touching and fingers spread, a thin sheet of
    flames shoots forth from your outstretched fingertips. Each creature in a
    15-foot cone must make a Dexterity saving throw. A creature takes 3d6 fire
    damage on a failed save, or half as much damage on a successful one.

    The fire
    ignites any flammable objects in the area that aren't being worn or carried.

    At
    Higher Levels: When you cast this spell using a spell slot of 2nd level or
    higher, the damage increases by 1d6 for each slot level above 1st.
    """

    name = "Burning Hands"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (15-foot cone)"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")



"""Registries for looking up pre-defined game content.

The function *find_content* will find the class for a piece of content
when given the name of that content. For example:
``find_content("leather armor")`` will return the
*dungeonsheets.armor.LeatherArmor* class.

The *find_content* function is a shortcut that makes use of the
*default_content_registry*, which is a *ContentRegistry* instance that
is aware of all the content included in *dungeonsheets*. **New modules
should register themselves** with the *default_content_registry*, best
achieved by::

    from content_registry import default_content_registry
    default_content_registry.add_module(__name__)

In the case of **homebrew content**, the python file may not be in the
python path and so cannot be imported directly. In this case, the
*import_homebrew* function will import the python file-name given and
then register it with the *default_content_registry*, most often
within a character sheet file. For example, if the *PaperSword* weapon
class is defined in a separate file "my_homebrew.py", then in the
character file::

    from content_registry import import_homebrew
    import_homebrew("my_homebrew.py")
    
    weapons = ["paper sword"]

If homebrew content shares a name with canonical content, then lookup
by string will raise an exception. In those situations, the homebrew
content can be used directly from *import_homebrew*::

    from content_registry import import_homebrew
    campaign = import_homebrew("my_homebrew.py")
    
    weapons = [campaign.PaperSword]

"""


import sys
from pathlib import Path
from functools import lru_cache
import importlib.util
from typing import Union, List, Optional

from dungeonsheets import exceptions


class ContentRegistry:
    modules = None

    def __init__(self):
        self.modules = []

    def add_module(self, new_module):
        """Register a module with this registry.

        Adding the same module multiple times has no effect.

        *new_module* can also be a string, in which case, an attempt
        will be made to load the module from *sys.modules*. This way,
        a module can register itself::

            # Define classes, etc
            ...
            # Register the module
            registry = ContentRegistry()
            registry.add_module(__name__)
        
        """
        # Try and look up the module by name
        try:
            new_module = sys.modules[new_module]
        except KeyError:
            if isinstance(new_module, str):
                raise exceptions.ContentNotFound(f"Module could not be resolved: {repr(new_module)}")
        # Add the imported module to the list for later
        if new_module not in self.modules:
            self.modules.append(new_module)

    def findattr(self, name, valid_classes=[]):
        """Resolve the name of a piece of content to the corresponding Class.

        Similar to builtin getattr(obj, name) but more forgiving to
        whitespace and capitalization.

        valid_classes
          If given, only subclasses of classes in this list will be
          returned.

        """
        # Come up with several options
        try:
            name = name.strip()
        except AttributeError as e:
            # Probably not a string
            raise ValueError('content "%s" is not a valid identifier string.' % name)
        # check for +X weapons, armor, shields
        bonus = 0
        for i in range(1, 11):
            if (f"+{i}" in name) or (f"+ {i}" in name):
                bonus = i
                name = name.replace(f"+{i}", "").replace(f"+ {i}", "")
                break
        py_name = (
            name.replace("-", "_").replace(" ", "_").replace("'", "").replace("/", "")
        )
        camel_case = "".join([s.capitalize() for s in py_name.split("_")])
        # Check each module in the registry
        found_attrs = []
        for module in self.modules:
            if hasattr(module, py_name):
                # Direct lookup
                found_attrs.append(getattr(module, py_name))
            elif hasattr(module, camel_case):
                # CamelCase lookup
                found_attrs.append(getattr(module, camel_case))
        # Filter by valid classes
        if len(valid_classes) > 0:
            is_valid = [False for attr in found_attrs]
            for cls in valid_classes:
                is_valid = [
                    v
                    or isinstance(attr, cls)
                    or (isinstance(attr, type) and issubclass(attr, cls))
                    for v, attr in zip(is_valid, found_attrs)
                ]
            found_attrs = [attr for attr, v in zip(found_attrs, is_valid) if v]
        # Check that we found a valid, unique attribute
        if len(found_attrs) == 0:
            raise exceptions.ContentNotFound(f"Modules {self.modules} have no attribute {name}")
        elif len(found_attrs) > 1:
            raise exceptions.AmbiguousContent(f"Found multiple content entries for {name}")
        else:
            attr = found_attrs[0]
        # Apply weapon/etc. bonuses
        if bonus > 0 and hasattr(attr, 'improved_version'):
            attr = attr.improved_version(bonus)
        return attr


default_content_registry = ContentRegistry()


def find_content(name: str, valid_classes: Optional[List] = None):
    """Find content from a previously registered module.

    Parameters
    ==========
    name
      The name of the item to find. Can be "CamelCase", "under_case",
      or "lower case".
    valid_classes
      A list of parent classes to look for. If ``None`` or ``[]``, all
      classes will be considered valid.

    """
    if valid_classes is None:
        valid_classes = []
    return default_content_registry.findattr(name, valid_classes=valid_classes)


@lru_cache()
def import_homebrew(filepath: Union[str, Path]):
    """Import a module file containing homebrew content.

    This is intended to be used in a character/GM sheet to load in
    homebrew content defined in an external file.

    Parameters
    ==========
    filepath
      The location of the python file containing the homebrew content.

    Returns
    =======
    mod
      The imported module of homebrew content.

    """
    spec = importlib.util.spec_from_file_location("module.name", filepath)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)
    default_content_registry.add_module(mod)
    return mod



from collections import defaultdict

from dungeonsheets import features, weapons
from dungeonsheets.classes.classes import CharClass, SubClass


# Eberron Rising from the Last War
class Alchemist(SubClass):
    """An Alchemist is an expert at combining reagents to produce mystical
    effects. Alchemists use their creations to give life and to leech it away.
    Alchemy is the oldest of artificer traditions, and its versatility has
    long been valued during times of war and peace.
    """

    name = "Alchemist"
    features_by_level = defaultdict(list)
    features_by_level[3] = [
        features.AlchemistToolProficiency,
        features.AlchemistSpells,
        features.ExperimentalElixir,
    ]
    features_by_level[5] = [features.AlchemicalSavant]
    features_by_level[9] = [features.RestorativeReagents]
    features_by_level[15] = [features.ChemicalMastery]


class Artillerist(SubClass):
    """An Artillerist specializes in using magic to hurl energy, projectiles,
    and explosions on a battlefield. This destructive power was valued by all
    the armies of the Last War. Now that the war is over, some members of this
    specialization have sought to build a more peaceful world by using their
    powers to fight the resurgence of strife in Khorvaire. The gnome artificer
    Vi, an unlikely yet key member of House Cannith's warforged project, has
    been especially vocal about making things right: "It's about time we fixed
    things instead of blowing them all to hell."
    """

    name = "Artillerist"
    features_by_level = defaultdict(list)
    features_by_level[3] = [
        features.ArtilleristSpells,
        features.ArtilleristToolProficiency,
        features.EldritchCannon,
    ]
    features_by_level[5] = [features.ArcaneFirearm]
    features_by_level[9] = [features.ExplosiveCannon]
    features_by_level[15] = [features.FortifiedPosition]


class BattleSmith(SubClass):
    """Armies require protection, and someone has to put things back together
    if defenses fail. A combination of protector and medic, a Battle Smith is
    an expert at defending others and repairing both material and personnel.
    To aid in their work, Battle Smiths are usually accompanied by a steel
    defender, a protective companion of their own creation. Many soldiers tell
    stories of nearly dying before being saved by a Battle Smith and a steel
    defender.

    Battle Smiths played a key role in House Cannith's work on battle
    constructs and the original warforged, and after the Last War, these
    artificers led efforts to aid those who were injured in the war's horrific
    battles.
    """

    name = "Battle Smith"
    features_by_level = defaultdict(list)
    features_by_level[3] = [
        features.BattleSmithSpells,
        features.BattleSmithToolProficiency,
        features.BattleReady,
        features.SteelDefender,
    ]
    features_by_level[5] = [features.ExtraAttackBattleSmith]
    features_by_level[9] = [features.ArcaneJolt]
    features_by_level[15] = [features.ImprovedDefender]


class Artificer(CharClass):
    name = "Artificer"
    hit_dice_faces = 8
    subclass_select_level = 3
    subclasses_available = (Alchemist, Artillerist, BattleSmith)
    saving_throw_proficiencies = ("intelligence", "constitution")
    primary_abilities = ("intelligence",)
    _proficiencies_text = (
        "Light armor",
        "Medium armor",
        "Shields",
        "Simple weapons",
        "Thieve's tools",
        "Tinker's tools",
        "One type of artisan's tools of your choice",
    )
    _multiclass_proficiencies_text = (
        "Light armor",
        "Medium armor",
        "Shields",
        "Thieve's tools",
        "Tinker's tools",
    )
    weapon_proficiencies = (weapons.SimpleWeapon,)
    infusions = []
    class_skill_choices = (
        "Arcana",
        "History",
        "Investigation",
        "Medicine",
        "Nature",
        "Perception",
        "Sleight of Hand",
    )
    features_by_level = defaultdict(list)
    features_by_level[1] = [
        features.MagicalTinkering,
        features.FirearmProficiency,
        features.ArtificerSpellcasting,
        features.ArtificerRitualCasting,
    ]
    features_by_level[2] = [features.InfuseItem]
    features_by_level[3] = [features.TheRightToolForTheJob]
    features_by_level[6] = [features.ToolExpertise]
    features_by_level[7] = [features.FlashOfGenius]
    features_by_level[10] = [features.MagicItemAdept]
    features_by_level[11] = [features.SpellStoringItem]
    features_by_level[14] = [features.MagicItemSavant]
    features_by_level[18] = [features.MagicItemMaster]
    features_by_level[20] = [features.SoulOfArtifice]
    spellcasting_ability = "intelligence"
    spell_slots_by_level = {
        1: (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        2: (2, 2, 0, 0, 0, 0, 0, 0, 0, 0),
        3: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        4: (2, 3, 0, 0, 0, 0, 0, 0, 0, 0),
        5: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        6: (2, 4, 2, 0, 0, 0, 0, 0, 0, 0),
        7: (2, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        8: (2, 4, 3, 0, 0, 0, 0, 0, 0, 0),
        9: (2, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        10: (3, 4, 3, 2, 0, 0, 0, 0, 0, 0),
        11: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        12: (3, 4, 3, 3, 0, 0, 0, 0, 0, 0),
        13: (3, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        14: (4, 4, 3, 3, 1, 0, 0, 0, 0, 0),
        15: (4, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        16: (4, 4, 3, 3, 2, 0, 0, 0, 0, 0),
        17: (4, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        18: (4, 4, 3, 3, 3, 1, 0, 0, 0, 0),
        19: (4, 4, 3, 3, 3, 2, 0, 0, 0, 0),
        20: (4, 4, 3, 3, 3, 2, 0, 0, 0, 0),
    }



from typing import Mapping
from html.parser import HTMLParser
import re
from pathlib import Path

from ebooklib import epub, ITEM_STYLE
from docutils import core
from sphinx.util.docstrings import prepare_docstring
from docutils.writers.html5_polyglot import Writer as HTMLWriter

from dungeonsheets.forms import dice_re, jinja_environment


def create_epub(
    chapters: Mapping, title: str, basename: str, use_dnd_decorations: bool = False
):
    """Prepare an EPUB file from the list of chapters.

    Parameters
    ==========
    chapters
      A mapping where the keys are chapter names (spines) and the
      values are strings of HTML to be rendered as the chapter
      contents.
    basename
      The basename for saving files (PDFs, etc). The resulting epub
      file will be "{basename}.epub".
    use_dnd_decorations
      If true, style sheets will be included to produce D&D stylized
      stat blocks, etc.

    """
    # Create a new epub book
    book = epub.EpubBook()
    book.set_identifier("id123456")
    book.set_title(title)
    book.set_language("en")
    # Add the css files
    css_template = jinja_env.get_template("dungeonsheets_epub.css")
    dl_widths = {  # Width for dl lists, in 'em' units
        "character-details": 11,
        "combat-stats": 15,
        "proficiencies": 8.5,
        "faction": 6,
        "spellcasting": 12.5,
        "spell-slots": 8,
        "spell-details": 10,
        "beast-stats": 9,
        "feature-details": 5.5,
        "infusion-details": 8.5,
        "magic-item-details": 13.5,
        "monster-details": 15,
    }
    style = css_template.render(use_dnd_decorations=use_dnd_decorations, dl_widths=dl_widths)
    css = epub.EpubItem(
        uid="style_default",
        file_name="style/gm_sheet.css",
        media_type="text/css",
        content=style,
    )
    book.add_item(css)
    # Add paper background
    with open(Path(__file__).parent / "forms/paper-low-res.jpg", mode="rb") as fp:
        bg_img = fp.read()
    paper = epub.EpubItem(
        file_name="images/paper.jpg",
        media_type="image/jpeg",
        content=bg_img,
    )
    book.add_item(paper)
    # Create the separate chapters
    toc = ["nav"]
    html_chapters = []
    for chap_title, content in chapters.items():
        chap_fname = chap_title.replace(" - ", "-").replace(" ", "_").lower()
        chap_fname = "{}.html".format(chap_fname)
        chapter = epub.EpubHtml(
            title=chap_title,
            file_name=chap_fname,
            lang="en",
            media_type="application/xhtml+xml",
        )
        chapter.set_content(content)
        chapter.add_item(css)
        book.add_item(chapter)
        html_chapters.append(chapter)
        # Add entries for the table of contents
        toc.append(
            toc_from_headings(
                html=content, filename=chap_fname, chapter_title=chap_title
            )
        )
    # Add the table of contents
    book.toc = toc
    book.spine = ("nav", *html_chapters)
    # add default NCX and Nav file
    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())
    # Save the file
    epub_fname = f"{basename}.epub"
    epub.write_epub(epub_fname, book)


class HeadingParser(HTMLParser):
    tag_re = re.compile("h(\d+)")
    _curr_level = None
    _curr_id = None
    _curr_title = None

    def __init__(self, *args, **kwargs):
        self.headings = []
        super().__init__(*args, **kwargs)

    def heading_level(self, tag):
        match = self.tag_re.match(tag)
        if match:
            return int(match.group(1))
        else:
            return None

    def handle_starttag(self, tag, attrs):
        this_level = self.heading_level(tag)
        if this_level is not None:
            # Found a heading, so process the properties
            self._curr_level = this_level
            attrs = {k: v for k, v in attrs}
            self._curr_id = attrs.get("id")

    def handle_endtag(self, tag):
        this_level = self.heading_level(tag)
        if this_level is not None and this_level == self._curr_level and self._curr_id is not None:
            heading = {
                "level": this_level,
                "id": self._curr_id,
                "title": self._curr_title,
            }
            self.headings.append(heading)
            # Reset the current values
            self._curr_level = None
            self._curr_id = None
            self._curr_title = None

    def handle_data(self, data):
        # Save the title
        if self._curr_level is not None:
            self._curr_title = data


def toc_from_headings(
    html: str, filename: str = "", chapter_title: str = "Sheet"
) -> list:
    """Accept a chapter of HTML, and extract a table of contents segment.

    Parameters
    ----------
    html
      The HTML block to be parsed.
    filename
      The name of this file to be used for hrefs. E.g.
      "index.html#heading_1".

    Returns
    -------
    toc
      A sequence of table-of-contents links.

    """
    # Parse the HTML
    parser = HeadingParser()
    parser.feed(html)
    headings = parser.headings
    # Parse into a table of contents
    if len(headings) == 0:
        # No headings found, so just the chapter link
        toc = epub.Link(href=filename, title=chapter_title, uid=filename)
    else:
        # Add a section for the chapter as a whole
        toc = (epub.Section(href=filename, title=chapter_title), [])
        sections_stack = [toc]
        # Parse all the headings
        for idx, heading in enumerate(headings):
            # Determine where we are in the tree
            href = f"{filename}#{heading['id']}"
            parent_section = sections_stack[-1]
            is_last = idx == (len(headings) - 1)
            is_leaf = is_last or heading["level"] >= headings[idx + 1]["level"]
            # Add a leaf or branch depending on the heading structure
            if is_leaf:
                parent_section[1].append(
                    epub.Link(href=href, title=heading["title"], uid=href.replace("#", ":"))
                )
            else:
                new_section = (epub.Section(href=href, title=heading["title"]), [])
                parent_section[1].append(new_section)
                sections_stack.append(new_section)
            # Walk back up the stack
            if not is_last:
                for idx in range(max(0, heading["level"] - headings[idx + 1]["level"])):
                    sections_stack.pop()

    return toc


def html_parts(
    input_string,
    source_path=None,
    destination_path=None,
    input_encoding="unicode",
    doctitle=True,
    initial_header_level=1,
):
    """
    Given an input string, returns a dictionary of HTML document parts.

    Dictionary keys are the names of parts, and values are Unicode strings;
    encoding is up to the client.

    Parameters:

    - `input_string`: A multi-line text string; required.
    - `source_path`: Path to the source file or object.  Optional, but useful
      for diagnostic output (system messages).
    - `destination_path`: Path to the file or object which will receive the
      output; optional.  Used for determining relative paths (stylesheets,
      source links, etc.).
    - `input_encoding`: The encoding of `input_string`.  If it is an encoded
      8-bit string, provide the correct encoding.  If it is a Unicode string,
      use "unicode", the default.
    - `doctitle`: Disable the promotion of a lone top-level section title to
      document title (and subsequent section title to document subtitle
      promotion); enabled by default.
    - `initial_header_level`: The initial level for header elements (e.g. 1
      for "<h1>").
    """
    # Remove indentation, etc
    input_string = "\n".join(prepare_docstring(input_string))
    # Parse from rst to TeX
    overrides = {
        "input_encoding": input_encoding,
        "doctitle_xform": doctitle,
        "initial_header_level": initial_header_level,
    }
    writer = HTMLWriter()
    parts = core.publish_parts(
        source=input_string,
        source_path=source_path,
        destination_path=destination_path,
        writer=writer,
        settings_overrides=overrides,
    )
    return parts


def rst_to_html(rst, top_heading_level: int = 0, format_dice: bool = True):
    """Basic markup of reST to HTML code.

    The translation between reST headings and LaTeX headings is
    modified by the *top_heading_level* parameter. A value of 0
    (default) translates "# Heading" -> "<h1>{Heading}</h1>". A value
    of 1 translates "# Heading" -> "<h2>{Heading}</h2>", etc.

    Note: heading translation is currently broken.

    Parameters
    ==========
    rst
      reStructured text input to be parsed.
    top_heading_level : optional
      The highest level heading that will be added to the HTML as
      described above.
    format_dice
      If true, dice strings (e.g. "1d4") will be formatted in
      monospace font.

    Returns
    =======
    html : str
      The reST text parsed into HTML markup.

    """
    if rst is None:
        # No reST, so return an empty string
        html = ""
    else:
        # Mark hit dice in monospace font
        if format_dice:
            rst = dice_re.sub(r"``\1``", rst)
        _html_parts = html_parts(rst)
        html = _html_parts["body"]
    return html


def to_heading_id(inpt: str) -> str:
    """Take a string and make it suitable for use as an HTML header id."""
    bad_characters = ["'", "(", ")", "&", "/", "+", ",", "=", ]
    for c in bad_characters:
        inpt = inpt.replace(c, "")
    return inpt.replace(" ", "-")



# Prepare the jinja environment
jinja_env = jinja_environment()
jinja_env.filters["rst_to_html"] = rst_to_html
jinja_env.filters["to_heading_id"] = to_heading_id



from pathlib import Path
import os
import re
import subprocess
import logging

from docutils import core
from docutils.writers.latex2e import Writer, Table, LaTeXTranslator
from sphinx.util.docstrings import prepare_docstring

from dungeonsheets import exceptions
from dungeonsheets.forms import dice_re


log = logging.getLogger(__name__)


class LatexWriter(Writer):
    def __init__(self):
        super().__init__()
        self.translator_class = DNDTranslator


class DNDTable(Table):
    pass


class DNDTranslator(LaTeXTranslator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings.table_style = ["borderless"]
        self.active_table = DNDTable(self, "supertabular")


def _remove_temp_files(basename_):
    # Convenience function for removing temporary files
    filenames = [
        Path(f"{basename_}.tex"),
        Path(f"{basename_}.aux"),
        Path(f"{basename_}.log"),
    ]
    for filename in filenames:
        if filename.exists():
            filename.unlink()


def create_latex_pdf(
    tex: str,
    basename: str,
    keep_temp_files: bool = False,
    use_dnd_decorations: bool = False,
):
    # Create tex document
    tex_file = f"{basename}.tex"
    with open(tex_file, mode="w", encoding="utf-8") as f:
        f.write(tex)

    # Compile the PDF
    pdf_file = Path(f"{basename}.pdf")
    output_dir = pdf_file.resolve().parent
    tex_command_line = [
        "pdflatex",
        "--output-directory",
        str(output_dir),
        "-halt-on-error",
        "-interaction=nonstopmode",
        str(tex_file),
    ]

    environment = os.environ
    tex_env = environment.get('TEXINPUTS', '')
    module_root = Path(__file__).parent / "modules/"
    module_dirs = [module_root / mdir for mdir in ["DND-5e-LaTeX-Template"]]
    log.debug(f"Loading additional modules from {module_dirs}.")
    texinputs = f".:{':'.join(str(d) for d in module_dirs)}:{module_root}:{tex_env}"
    environment['TEXINPUTS'] = texinputs
    passes = 2 if use_dnd_decorations else 1
    log.debug(tex_command_line)
    log.debug("LaTeX command: %s" % " ".join(tex_command_line))
    log.debug("LaTeX TEXINPUTS: %s" % texinputs)
    log.debug("LaTeX environ:")
    for key, val in environment.items():
        log.debug("    %s: %s" % (key, val))
    try:
        for i in range(passes):
            result = subprocess.run(
                tex_command_line, stdout=subprocess.DEVNULL, env=environment, timeout=30
            )
    except FileNotFoundError:
        # Remove temporary files
        _remove_temp_files(basename)
        raise exceptions.LatexNotFoundError()
    else:
        if result.returncode == 0 and not keep_temp_files:
            _remove_temp_files(basename)
        if result.returncode != 0:
            # Prepare to raise an exception
            logfile = Path(f"{basename}.log")
            err_msg = f"Processing of {basename}.tex failed. See {logfile} for details."
            # Load the log file for more details
            tex_error_msg = tex_error(logfile)
            if tex_error_msg:
                for line in tex_error_msg.split("\n"):
                    log.error(line)
            raise exceptions.LatexError(err_msg)
    finally:
        environment['TEXINPUTS'] = tex_env


def tex_error(logfile: Path) -> str:
    """Parse a LaTeX log file and look for errors."""
    has_error = False
    error_lines = []
    if logfile.exists():
        with open(logfile, mode="r") as fp:
            for line in fp.readlines():
                # Check for the start of an error message
                if "LaTeX Error" in line:
                    has_error = True
                # We've already found an error, so save this line for later
                if has_error:
                    error_lines.append(line)
    return "".join(error_lines)


def latex_parts(
    input_string,
    source_path=None,
    destination_path=None,
    input_encoding="unicode",
    doctitle=True,
    initial_header_level=1,
):
    """
    Given an input string, returns a dictionary of HTML document parts.

    Dictionary keys are the names of parts, and values are Unicode strings;
    encoding is up to the client.

    Parameters:

    - `input_string`: A multi-line text string; required.
    - `source_path`: Path to the source file or object.  Optional, but useful
      for diagnostic output (system messages).
    - `destination_path`: Path to the file or object which will receive the
      output; optional.  Used for determining relative paths (stylesheets,
      source links, etc.).
    - `input_encoding`: The encoding of `input_string`.  If it is an encoded
      8-bit string, provide the correct encoding.  If it is a Unicode string,
      use "unicode", the default.
    - `doctitle`: Disable the promotion of a lone top-level section title to
      document title (and subsequent section title to document subtitle
      promotion); enabled by default.
    - `initial_header_level`: The initial level for header elements (e.g. 1
      for "<h1>").
    """
    # Remove indentation, etc
    input_string = "\n".join(prepare_docstring(input_string))
    # Parse from rst to TeX
    overrides = {
        "input_encoding": input_encoding,
        "doctitle_xform": doctitle,
        "initial_header_level": initial_header_level,
    }
    writer = LatexWriter()
    parts = core.publish_parts(
        source=input_string,
        source_path=source_path,
        destination_path=destination_path,
        writer=writer,
        settings_overrides=overrides,
    )
    return parts


def rst_to_latex(rst, top_heading_level: int=0, format_dice: bool = True, use_dnd_decorations=False):
    """Basic markup of reST to LaTeX code.

    The translation between reST headings and LaTeX headings is
    modified by the *top_heading_level* parameter. A value of 0
    (default) translates "# Heading" -> "\\section{Heading}". A value
    of 1 translates "# Heading" -> "\\subsection{Heading}", etc.

    Note: heading translation is currently broken.

    Parameters
    ==========
    rst
      reStructured text input to be parsed.
    top_heading_level : optional
      The highest level heading that will be added to the LaTeX as
      described above.
    format_dice
      If true, dice strings (e.g. "1d4") will be formatted in
      monospace font.
    
    Returns
    =======
    tex : str
      The reST text parsed into LaTeX markup.

    """
    if rst is None:
        # No reST, so return an empty string
        tex = ""
    else:
        # Mark hit dice in monospace font
        if format_dice:
            rst = dice_re.sub(r"``\1``", rst)
        tex_parts = latex_parts(rst)
        tex = tex_parts["body"]
    # Apply fancy D&D decorations
    if use_dnd_decorations:
        tex = re.sub(r"p{[0-9.]+\\DUtablewidth}", "l ", tex, flags=re.M)
        tex = tex.replace(r"\begin{supertabular}[c]", r"\begin{DndTable}")
        tex = tex.replace(r"\end{supertabular}", r"\end{DndTable}")
    return tex



from dungeonsheets.spells.spells import Spell


class CallLightning(Spell):
    """A storm cloud appears in the shape of a cylinder that is 10 feet
    tall with a 60-foot radius, centered on a point you can see 100
    feet directly above you. The spell fails if you can't see a point
    in the air where the storm cloud could appear (for example, if you
    are in a room that can't accommodate the cloud).

    When you cast the spell, choose a point you can see within
    range. A bolt of lightning flashes down from the cloud to that
    point. Each creature within 5 feet of that point must make a
    Dexterity saving throw. A creature takes 3d10 lightning damage on
    a failed save, or half as much damage on a successful one.  On
    each of your turns until the spell ends, you can use your action
    to call down lightning in this way again, targeting the same point
    or a different one.

    If you are outdoors in stormy conditions when you cast this spell,
    the spell gives you control over the existing storm instead of
    creating a new one. Under such conditions, the spell's damage
    increases by 1d10.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 4th or higher level, the damage increases by 1d10 for each slot
    level above 3rd.

    """

    name = "Call Lightning"
    level = 3
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid",)


class CalmEmotions(Spell):
    """You attempt to suppress strong emotions in a group of people.  Each
    humanoid in a 20-foot-radius sphere centered on a point you choose
    within range must make a Charisma saving throw; a creature can
    choose to fail this saving throw if it wishes. If a creature fails
    its saving throw, choose one of the following two effects. You can
    suppress any effect causing a target to be charmed or
    frightened. When this spell ends, any suppressed effect resumes,
    provided that its duration has not expired in the meantime.

    Alternatively, you can make a target indifferent about creatures
    of your choice that it is hostile toward.  This indifference ends
    if the target is attacked or harmed by a spell or if it witnesses
    any of its friends being harmed. When the spell ends, the creature
    becomes hostile again, unless the DM rules otherwise.

    """

    name = "Calm Emotions"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Cleric")


class Catapult(Spell):
    """Choose one object weighing 1 to 5 pounds within range that isn't
    being worn or carried. The object flies in a straight line up to
    90 feet in a direction you choose before falling to the ground,
    stopping early if it impacts against a solid surface. If the
    object would strike a creature, that creature must make a
    Dexterity saving throw. On a failed save, the object strikes the
    target and stops moving. In either case, both the object and the
    creature or solid surface take 3d8 bludgeoning damage.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 2nd level or higher, the maximum weight of objects that you can
    target with this spell increases by 5 pounds, and the damage
    increases by 1d8, for each slot level above 1st.

    """

    name = "Catapult"
    level = 1
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("S",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class Catnap(Spell):
    """You make a calming gesture, and up to three willing creatures of
    your choice that you can see within range fall unconscious for the
    spell's duration. The spell ends on a target early if it takes
    damage or someone uses an action to shake or slap it awake. If a
    target remains unconscious for the full duration, that target
    gains the benefit of a short rest, and it can't be affected by
    this spell again until it finishes a long rest.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 4th level or higher, you can target one additional willing
    creature for each slot level above 3rd.

    """

    name = "Catnap"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("S", "M")
    materials = "A pinch of sand"
    duration = "10 minutes"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Wizard", "Bard", "Sorcerer")


class CauseFear(Spell):
    """You awaken the sense of mortality in one creature you can see within range. A
    construct or an undead is immune to this effect. The target must succeed on a
    Wisdom saving throw or become frightened of you until the spell ends. The
    frightened target can repeat the saving throw at the end of each of its turns,
    ending the effect on itself on a success.

    **At Higher Levels:** When you cast this
    spell using a spell slot of 2nd level or higher, you can target one additional
    creature for each slot level above lst. The creatures must be within 30 feet of
    each other when you target them.
    """

    name = "Cause Fear"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Warlock", "Wizard")


class Ceremony(Spell):
    """You perform a special religious ceremony that is infused with magic. When you
    cast the spell, choose one of the following rites, the target of which must be
    within 10 feet of you throughout the casting.
    Atonement. You touch one willing
    creature whose alignment has changed, and you make a DC 20 Wisdom (Insight)
    check. On a successful check, you restore the target to its original alignment.

    Bless Water. You touch one vial of water and cause it to become holy water.

    Coming of Age. You touch one humanoid who is a young adult. For the next 24
    hours, whenever the target makes an ability check, it can roll a d4 and add the
    number rolled to the ability check. A creature can benefit from this rite only
    once.
    Dedication. You touch one humanoid who wishes to be dedicated to your
    god's service. For the next 24 hours, whenever the target makes a saving throw,
    it can roll a d4 and add the number rolled to the save. A creature can benefit
    from this rite only once.
    Funeral Rite. You touch one corpse, and for the next 7
    days, the target can't become undead by any means short of a wish spell.

    Wedding. You touch adult humanoids willing to be bonded together in marriage.
    For the next 7 days, each target gains a +2 bonus to AC while they are within 30
    feet of each other. A creature can benefit from this rite again only if
    widowed.
    """

    name = "Ceremony"
    level = 1
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "25 gp worth of powdered silver,which the spell consumes"
    duration = "Instantaneous"
    ritual = True
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin")


class ChainLightning(Spell):
    """You create a bolt of lightning that arcs toward a target of your choice that you
    can see within range. Three bolts then leap from that target to as many as
    three other targets, each of which must be within 30 feet of the first target. A
    target can be a creature or an object and can be targeted by only one of the
    bolts.

    A target must make a Dexterity saving throw. The target takes 10d8
    lightning damage on a failed save, or half as much on a successful one.

    At
    Higher Levels: When you cast this spell using a spell slot of 7th level or
    higher, one additional bolt leaps from the first target to another target for
    each slot level above 6th.
    """

    name = "Chain Lightning"
    level = 6
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = (
        "A bit of fur; a piece of amber, glass, or a crystal rod; and three silver pins"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class ChaosBolt(Spell):
    """You hurl an undulating, warbling mass of chaotic energy at one
    creature in range. Make a ranged spell attack against the
    target. On a hit, the target takes ``2d8+1d6`` damage. Choose one
    of the d8s. The number rolled on that die determines the attacks
    damage type, as shown below.

    == ===========
    d8 Damage Type
    == ===========
    1  Acid
    2  Cold
    3  Fire
    4  Force
    5  Lightning
    6  Poison
    7  Psychic
    8  Thunder
    == ===========

    If you roll the same number on both d8s, the chaotic energy leaps
    from the target to a different creature of your choice within 30
    feet of it. Make a new attack roll against the new target, and
    make a new damage roll, which could cause the chaotic energy to
    leap again. A creature can be targeted only once by each casting
    of this spell.

    **At Higher Levels:** When you cast this spell using a spell slot of
    2nd level or higher, each target takes 1d6 extra damage of the
    type rolled for each slot level above 1st.

    """

    name = "Chaos Bolt"
    level = 1
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer",)


class CharmMonster(Spell):
    """You attempt to charm a creature you can see within range. It must
    make a Wisdom saving throw, and it does so with advantage if you
    or your companions are fighting it. If it fails the saving throw,
    it is charmed by you until the spell ends or until you or your
    companions do anything harmful to it. The charmed creature is
    friendly to you. When the spell ends, the creature knows it was
    charmed by you.

    **At Higher Levels:** When you cast this spell using a spell slot of
    5th level or higher, you can target one additional creature for
    each slot level above 4th. The creatures must be within 30 feet of
    each other when you target them.

    """

    name = "Charm Monster"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Druid", "Sorcerer", "Warlock", "Wizard")


class CharmPerson(Spell):
    """You attempt to charm a humanoid you can see within range.  It must
    make a Wisdom saving throw, and does so with advantage if you or
    your companions are fighting it. If it fails the saving throw, it
    is charmed by you until the spell ends or until you or your
    companions do anything harmful to it.The charmed creature regards
    you as a friendly acquaintance. When the spell ends, the creature
    knows it was charmed by you.

    **At Higher Levels:** When you cast this spell using a spell slot of
    2nd level or higher, you can target one additional creature for
    each slot level above 1st. The creatures must be within 30 feet of
    each other when you target them.

    """

    name = "Charm Person"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Druid", "Sorcerer", "Warlock", "Wizard")


class ChillTouch(Spell):
    """You create a ghostly, skeletal hand in the space of a creature within range.

    Make a ranged spell attack against the creature to assail it with the chill of
    the grave. On a hit, the target takes 1d8 necrotic damage, and it can't regain
    hit points until the start of your next turn. Until then, the hand clings to the
    target. If you hit an undead target, it also has disadvantage on attack rolls
    against you until the end of your next turn.

    **At Higher Levels:** This spell's
    damage increases by 1d8 when you reach 5th level (2d8), 11th level (3d8), and
    17th level (4d8).
    """

    name = "Chill Touch"
    level = 0
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 round"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Sorcerer", "Warlock", "Wizard")


class ChromaticOrb(Spell):
    """You hurl a 4-inch-diameter sphere of energy at a creature that you can see
    within range. You choose acid, cold, fire, lightning, poison, or thunder for the
    type of orb you create, and then make a ranged spell attack against the target.
    If the attack hits, the creature takes 3d8 damage of the type you chose.

    At
    Higher Levels: When you cast this spell using a spell slot of 2nd level or
    higher, the damage increases by 1d8 for each slot level above 1st.
    """

    name = "Chromatic Orb"
    level = 1
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "A diamond worth at least 50 gp"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class CircleOfDeath(Spell):
    """A sphere of negative energy ripples out in a 60-foot-radius sphere from a point
    within range. Each creature in that area must make a Constitution saving throw.
    A target takes 8d6 necrotic damage on a failed save, or half as much damage on a
    successful one.

    **At Higher Levels:** W hen you cast this spell using a spell slot
    of 7th level or higher, the damage increases by 2d6 for each slot level above
    6th.
    """

    name = "Circle Of Death"
    level = 6
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = "The powder of a crushed black pearl worth at least 500 gp"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Sorcerer", "Warlock", "Wizard")


class CircleOfPower(Spell):
    """Divine energy radiates from you, distorting and diffusing magical energy within
    30 feet of you.
    Until the spell ends, the sphere moves with you, centered on
    you. For the duration, each friendly creature in the area (including you) has
    advantage on saving throws against spells and other magical effects.


    Additionally, when an affected creature succeeds on a saving throw made against
    a spell or magical effect that allows it to make a saving throw to take only
    half damage, it instead takes no damage if it succeeds on the saving throws.
    """

    name = "Circle Of Power"
    level = 5
    casting_time = "1 action"
    casting_range = "Self (30-foot radius)"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Paladin",)


class Clairvoyance(Spell):
    """You create an invisible sensor within range in a location familiar to you (a
    place you have visited or seen before) or in an obvious location that is
    unfamiliar to you (such as behind a door, around a corner, or in a grove of
    trees). The sensor remains in place for the duration, and it can't be attacked
    or otherwise interacted with.

    When you cast the spell, you choose seeing or
    hearing. You can use the chosen sense through the sensor as if you were in its
    space. As your action, you can switch between seeing and hearing. A creature
    that can see the sensor (such as a creature benefitting from see invisibility or
    truesight) sees a luminous, intangible orb about the size of your fist.
    """

    name = "Clairvoyance"
    level = 3
    casting_time = "10 minutes"
    casting_range = "1 mile"
    components = ("V", "S", "M")
    materials = (
        "A focus worth at least 100 gp, either a jeweled horn for hearing or a glass"
        " eye for seeing"
    )
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Divination"
    classes = ("Bard", "Cleric", "Sorcerer", "Wizard")


class Clone(Spell):
    """This spell grows an inert duplicate of a living creature as a safeguard against
    death.
    This clone forms inside a sealed vessel and grows to full size and
    maturity after 120 days; you can also choose to have the clone be a younger
    version of the same creature. It remains inert and endures indefinitely, as long
    as its vessel remains undisturbed.


     At any time after the clone matures, if
    the original creature dies, its soul transfers to the clone, provided that the
    soul is free and willing to return. The clone is physically identical to the
    original and has the same personality, memories, and abilities, but none of the
    original's equipment. The original creature's physical remains, if they still
    exist, becom e inert and can't thereafter be restored to life, since the
    creature's soul is elsewhere.
    """

    name = "Clone"
    level = 8
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "A diamond worth at least 1,000 gp and at least 1 cubic inch of flesh of the"
        " creature that is to be cloned, which the spell consum es, and a vessel worth"
        " at least 2,000 gp that has a sealable lid and is large enough to hold a"
        " medium creature, such"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Wizard",)


class CloudOfDaggers(Spell):
    """You fill the air with spinning daggers in a cube 5 feet on each side, centered
    on a point you choose within range. A creature takes 4d4 slashing damage when it
    enters the spell's area for the first time on a turn or starts its turn there.


    **At Higher Levels:** When you cast this spell using a spell slot of 3rd level or
    higher, the damage increases by 2d4 for each slot level above 2nd.
    """

    name = "Cloud Of Daggers"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A sliver of glass"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class Cloudkill(Spell):
    """You create a 20-foot-radius sphere of poisonous, yellow-green fog centered on a
    point you choose within range. The fog spreads around corners. It lasts for the
    duration or until strong wind dispereses the fog, ending the spell. Its area is
    heavily obscured.

    When a creature enters the spell's area for the first time
    on a turn or starts its turn there, that creature must make a Constitution
    saving throw. The creature takes 5d8 poison damageon a failed save, or half as
    much damage on a successful one. Creatures are affected even if they hold their
    breath or don't need to breathe.

    The fog moves 10 feet away from you at the
    start of each of your turns, rolling along the surface of the ground. The
    vapors, being heavier than air, sink to the lowest level of the land, even
    pouring down openings.

    **At Higher Levels:** When you cast this spell using a spell
    slot of 6th level or higher, the damage increases by 1d8 for each slot level
    above 5th.
    """

    name = "Cloudkill"
    level = 5
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Sorcerer", "Wizard")


class ColorSpray(Spell):
    """A dazzling array of flashing, colored light springs from your hand.
    Roll 6d10,
    the total is how many hit points of creatures this spell can effect. Creatures
    in a 15-foot cone originating from you are affected in ascending order of their
    current hit points (ignoring unconscious creatures and creatures that can't
    see).

    Starting with the creature that has the lowest current hit points, each
    creature affected by this spell is blinded until the spell ends. Subtract each
    creature's hit points from the total before moving on to the creature with the
    next lowest hit points. A creature's hit points must be equal to or less than
    the remaining total for the creature to be affected.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 2nd level or higher, roll an additional
    2d10 for each slot level above 1st.
    """

    name = "Color Spray"
    level = 1
    casting_time = "1 action"
    casting_range = "Self (15-foot cone)"
    components = ("V", "S", "M")
    materials = ""
    duration = "1 round"
    ritual = False
    magic_school = "Illusion"
    classes = ("Sorcerer", "Wizard")


class Command(Spell):
    """You speak a one-word command to a creature you can see within
    range.  The target must succeed on a Wisdom saving throw or follow
    the command on its next turn.  The spell has no effect if the
    target is undead, if it doesn't understand your language, or if
    your command is directly harmful to it.

    Some typical commands and their effects follow. You might issue a
    command other than one described here. If you do so, the DM
    determines how the target behaves. If the target can't follow your
    command, the spell ends.

    Approach.
      The target moves toward you by the shortest and most direct
      route, ending its turn if it moves within 5 feet of you.
    Drop.
      The target drops whatever it is holding and then ends its turn.
    Flee.
      The target spends its turn moving away from you by the fastest
      available means.
    Grovel.
      The target falls prone and then ends its turn.
    Halt.
      The target doesn't move and takes no actions. A flying creature
      stays aloft, provided that it is able to do so. If it must move
      to stay aloft, it flies the minimum distance needed to remain in
      the air.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 2nd level or higher, you can affect one additional creature for
    each slot level above 1st. The creatures must be within 30 feet of
    each other when you target them

    """

    name = "Command"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "1 round"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Cleric", "Paladin")


class Commune(Spell):
    """You contact your deity or a divine proxy and ask up to three questions that can
    be answered with a yes or no. You must ask your questions before the spell ends.
    You receive a correct answer for each question.

    Divine beings aren't
    necessarily omniscient, so you might receive "unclear" as an answer if a
    question pertains to information that lies beyond the deity's knowledge. In a
    case where a one-word answer could be misleading or contrary to the deity's
    interests, the DM might offer a short phrase as an answer instead.

    If you cast
    the spell two or more times before finishing your next long rest, there is a
    cumulative 25 percent chance for each casting after the first that you get no
    answer. The DM makes this roll in secret.
    """

    name = "Commune"
    level = 5
    casting_time = "1 minute"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "Incense and a vial of holy or unholy water"
    duration = "1 minute"
    ritual = True
    magic_school = "Divination"
    classes = ("Cleric",)


class CommuneWithNature(Spell):
    """You briefly become one with nature and gain knowledge of the
    surrounding territory.  In the outdoors, the spell gives you
    knowledge of the land within 3 miles of you. In caves and other
    natural underground settings, the radius is limited to 300
    feet. The spell doesn't function where nature has been replaced by
    construction, such as in dungeons and towns.

    You instantly gain knowledge of up to three facts of your choice
    about any of the following subjects as they relate to the area:

    - terrain and bodies of water
    - prevalent plants, minerals, animals, or peoples
    - powerful celestials, fey, fiends, elementals, or undead
    - influence from other planes of existence
    - buildings

    For example, you could determine the location of powerful undead
    in the area, the location of major sources of safe drinking water,
    and the location of any nearby towns.

    """

    name = "Commune With Nature"
    level = 5
    casting_time = "1 minute"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = True
    magic_school = "Divination"
    classes = ("Druid", "Ranger")


class CompelledDuel(Spell):
    """You attempt to compel a creature into a duel.
    One creature that you can see
    within range must make a Wisdom saving throw. On a failed save, the creature is
    drawn to you, compelled by your divine demand. For the duration, it has
    disadvantage on attack rolls against creatures other than you, and must make a
    Wisdom saving throw each time it attempts to move to a space that is more than
    30 feet away from you; if it succeeds on this saving throw, this spell doesn't
    restrict the target's movement for that turn.

    The spell ends if you attack any
    other creature, if you cast a spell that targets a hostile creature other than
    the target, if a creature friendly to you damages the target or casts a harmful
    spell on it, or if you end your turn more than 30 feet away from the target.
    """

    name = "Compelled Duel"
    level = 1
    casting_time = "1 bonus action"
    casting_range = "30 feet"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Paladin",)


class ComprehendLanguages(Spell):
    """For the duration, you understand the literal meaning of any spoken language that
    you hear.
    You also understand any spoken language that you hear. You also
    understand any written language that you see, but you must be touching the
    surface of which the words are written. It takes about 1 minute to read one page
    of text.

    This spell doesn't decode secret messages in a text or glyph, such
    as an arcane sigil, that isn't part of a written language.
    """

    name = "Comprehend Languages"
    level = 1
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A pinch of soot and salt"
    duration = "1 hour"
    ritual = True
    magic_school = "Divination"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class Compulsion(Spell):
    """Creatures of your choice that you can see within range and that can hear you
    must make a Wisdom saving throw.
    A target automatically succeeds on this saving
    throw if it can't be charmed. On a failed save, a target is affected by this
    spell. Until the spell ends, you can use a bonus action on each of your turns to
    designate a direction that is horizontal to you. Each affected target must use
    as much of its movement as possible to move in that direction on its next turn.
    It can take its action before it moves. After moving in this way, it can make
    another Wisdom saving throw to try to end the effect.

    A target isn't compelled
    to move into an obviously deadly hazard, such as a fire pit, but it will
    provoke opportunity attacks to move in the designated direction.
    """

    name = "Compulsion"
    level = 4
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard",)


class ConeOfCold(Spell):
    """A blast of cold air erupts from your hands.
    Each creature in a 60-foot cone
    must make a Constitution saving throw.

    A creature takes 8d8 cold damage on a
    failed save, or half as much damage on a successful one. A creature killed by
    this spell becomes a frozen statue until it thaws.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 6th level or higher, the damage increases
    by 1d8 for each slot level above 5th.
    """

    name = "Cone Of Cold"
    level = 5
    casting_time = "1 action"
    casting_range = "Self (60-foot cone)"
    components = ("V", "S", "M")
    materials = "A small crystal or glass cone"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class Confusion(Spell):
    """This spell assaults and twists creatures' minds, spawning delusions
    and provoking uncontrolled actions. Each creature in a
    10-foot-radius sphere centered on a point you choose within range
    must succeed on a Wisdom saving throw when you cast this spell or
    be affected by it.

    An affected target can't take reactions and must roll a d10 at the
    start of each of its turns to determine its behavior for that
    turn.

    d10 behavior:

    **1.** The creature uses all its movement to move in a random
    direction. To determine the direction, roll a d8 and assign a
    direction to each die face. The creature doesn't take an action
    this turn.

    **2-6.** The creature doesn't move or take actions this turn.

    **7-8.** The creature uses its action to make a melee attack
    against a randomly determined creature within its reach. If there
    is no creature within its reach, the creature does nothing this
    turn.

    **9-10.** The creature can act and move normally.

    At the end of its turns, an affected target can make a Wisdom
    saving throw. It it succeeds, this effect ends for that target.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 5th level or higher, the radius of the sphere increases by 5
    feet for each slot level above 4th

    """

    name = "Confusion"
    level = 4
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "Three nut shells"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Druid", "Sorcerer", "Wizard")


class ConjureAnimals(Spell):
    """You summon fey spirits that take the form of beasts and appear in
    unoccupied spaces that you can see within range.

    Choose one of the following options for what appears:

    - One beast of challenge rating 2 or lower
    - Two beasts of challenge rating 1 or lower
    - Four beasts of challenge rating 1/2 or lower
    - Eight beasts of challenge rating 1/4 or lower

    Each beast is also considered fey, and it disappears when it drops
    to 0 hit points or when the spell ends.


    The summoned creatures are friendly to you and your
    companions. Roll initiative for the summoned creatures as a group,
    which has its own turns. They obey any verbal commands that you
    issue to them (no action required by you). If you don't issue any
    commands to them, they defend themselves from hostile creatures,
    but otherwise take no actions.  The DM has the creatures'
    statistics.

    **At Higher Levels:** When you cast this spell using certain
    higher-level spell slots, you choose one of the summoning options
    above, and more creatures appear:

    - twice as many with a 5th-level slot
    - three times as many with a 7th-level slot
    - four times as many with a 9th-level slot.

    """

    name = "Conjure Animals"
    level = 3
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Ranger")


class ConjureBarrage(Spell):
    """You throw a nonmagical weapon or fire a piece of nonmagical ammunition into the
    air to create a cone of identical weapons that shoot forward and then disappear.
    Each creature in a 60-foot cone must succeed on a Dexterity saving throw. A
    creature takes 3d8 damage on a failed save, or half as much damage on a
    successful one. The damage type is the same as that of the weapon or ammunition
    used as a component.
    """

    name = "Conjure Barrage"
    level = 3
    casting_time = "1 action"
    casting_range = "Self (60-foot cone)"
    components = ("V", "S", "M")
    materials = "One piece of ammunition or a thrown weapon"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Ranger",)


class ConjureCelestial(Spell):
    """You summon a celestial of challenge rating 4 or lower, which appears in an
    unoccupied space that you can see within range. The celestial disappears when it
    drops to 0 hit points or when the spell ends.

    The celestial is friendly to
    you and your companions for the duration. Roll initiative for the celestial,
    which has its own turns. It obeys any verbal commands that you issue to it (no
    action required by you), as long as they don't violate its alignment. If you
    don't issue any commands to the celestial, it defends itself from hostile
    creatures but otherwise takes no actions
    The DM has the celestial's statistics.


    **At Higher Levels:** When you cast this spell using a 9th-level spell slot, you
    summon a celestial of challenge rating 5 or lower.
    """

    name = "Conjure Celestial"
    level = 7
    casting_time = "1 minute"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric",)


class ConjureElemental(Spell):
    """You call forth an elemental servant.
    Choose an area of air, earth, fire, or
    water that fills a 10-foot cube within range. An elemental of challenge rating 5
    or lower appropriate to the area you chose appears in an unoccupied space
    within 10 feet of it. For example, a fire elemental emerges from a bonfire, and
    an earth elemental rises up from the ground. The elemental disappears when it
    drops to 0 hit points or when the spell ends.

    The elemental is friendly to you
    and your companions for the duration. Roll initiative for the elemental, which
    has its own turns. It obeys any verbal commands that you issue to it (no action
    required by you). If you don't issue any commands to the elemental, it defends
    itself from hostile creatures but otherwise takes no actions.

    If your
    concentration is broken, the elemental doesn't disappear. Instead, you lose
    control of the elemental, it becom es hostile toward you and your companions,
    and it might attack. An uncontrolled elemental can't be dismissed by you, and it
    disappears 1 hour after you summoned it. The DM has the elemental's statistics.


    **At Higher Levels:** When you cast this spell using a spell slot of 6th level or
    higher, the challenge rating increases by 1 for each slot level above 5th.
    """

    name = "Conjure Elemental"
    level = 5
    casting_time = "1 minute"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = (
        "Burning incense for air, soft clay for earth, sulfur and phosphorus for fire,"
        " or water and sand for water"
    )
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Wizard")


class ConjureFey(Spell):
    """You summon a fey creature of challenge rating 6 or lower, or a fey spirit that
    takes the form of a beast of challenge rating 6 or lower.
    It appears in an
    unoccupied space that you can see within range. The fey creature disappears when
    it drops to 0 hit points or when the spell ends.

    The fey creature is friendly
    to you and your companions for the duration. Roll initiative for the creature,
    which has its own turns. It obeys any verbal commands that you issue to it (no
    action required by you), as long as they don't violate its alignment. If you
    don't issue any commands to the fey creature, it defends itself from hostile
    creatures but otherwise takes no actions.

    If your concentration is broken, the
    fey creature doesn't disappear. Instead, you lose control of the fey creature,
    it becomes hostile toward you and your companions, and it might attack. An
    uncontrolled fey creature can't be dismissed by you, and it disappears 1 hour
    after you summoned it.
    The DM has the fey creature's statistics.

    At Higher
    Levels: When you cast this spell using a spell slot of 7th level or higher, the
    challenge rating increases by 1 for each slot level above 6th
    """

    name = "Conjure Fey"
    level = 6
    casting_time = "1 minute"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Warlock")


class ConjureMinorElementals(Spell):
    """You summon elementals that appear in unoccupied spaces that you can see within
    range.
     You choose one the following options for what appears:
    -One elemental
    of challenge rating 2 or lower
    -Two elementals of challenge rating 1 or lower

    -Four elementals of challenge rating 1/2 or lower
    -Eight elementals of
    challenge rating 1/4 or lower.

    An elemental summoned by this spell disappears
    when it drops to 0 hit points or when the spell ends.

    The summoned creatures
    are friendly to you and your companions. Roll initiative for the summoned
    creatures as a group, which has its own turns. They obey any verbal commands
    that you issue to them (no action required by you). If you don't issue any
    commands to them, they defend themselves from hostile creatures, but otherwise
    take no actions.
    The DM has the creatures' statistics.

    **At Higher Levels:** When
    you cast this spell using certain higher-level spell slots, you choose one of
    the summoning options above, and more creatures appear: twice as many with a
    6th-level slot and three times as many with an 8th-level slot.
    """

    name = "Conjure Minor Elementals"
    level = 4
    casting_time = "1 minute"
    casting_range = "90 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Wizard")


class ConjureVolley(Spell):
    """You fire a piece of nonmagical ammunition from a ranged weapon or throw a
    nonmagical weapon into the air and choose a point within range.
    Hundreds of
    duplicates of the ammunition or weapon fall in a volley from above and then
    disappear. Each creature in a 40-foot-radius. 20-foot-high cylinder centered on
    that point must make a Dexterity saving throw. A creature takes 8d8 damage on a
    failed save, or half as much damage on a successful one. The damage type is the
    same as that of the ammunition or weapon.
    """

    name = "Conjure Volley"
    level = 5
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = "One piece of ammunition or one thrown weapon"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Ranger",)


class ConjureWoodlandBeings(Spell):
    """You summon fey creatures that appear in unoccupied spaces that you can see
    within range.

    Choose one of the following options for what appears:
     -  One
    fey creature of challenge rating 2 or lower
     -  Two fey creatures of challenge
    rating 1 or lower
     -  Four fey creatures of challenge rating 1/2 or lower
     -
    Eight fey creatures of challenge rating 1/4 or lower

    A summoned creature
    disappears when it drops to 0 hit points or when the spell ends.

    The summoned
    creatures are friendly to you and your companions. Roll initiative for the
    summoned creatures as a group, which have their own turns. They obey any verbal
    commands that you issue to them (no action required by you). If you don't issue
    any commands to them, they defend themselves from hostile creatures, but
    otherwise take no actions.
    The DM has the creatures' statistics.

    At Higher
    Levels: When you cast this spell using certain higher-level spell slots, you
    choose one of the summoning options above, and more creatures appear:
    twice as
    many with a 6th-level slot
    three times as many with an 8th-level slot.
    """

    name = "Conjure Woodland Beings"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "One holly berry per creature summoned"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Ranger")


class ContactOtherPlane(Spell):
    """You mentally contact a demigod, the spirit of a long-dead sage, or some other
    mysterious entity from another plane.
    Contacting this extraplanar intelligence
    can strain or even break your mind. When you cast this spell, make a DC 15
    Intelligence saving throw. On a failure, you take 6d6 psychic damage and are
    insane until you finish a long rest. While insane, you can't take actions, can't
    understand what other creatures say, can't read, and speak only in gibberish. A
    greater restoration spell cast on you ends this effect.

    On a successful save,
    you can ask the entity up to five questions. You must ask your questions before
    the spell ends. The DM answers each question with one word, such as yes,
    no, maybe, never, irrelevant, or unclear (if the entity doesn't know
    the answer to the question). If a one-word answer would be misleading, the DM
    might instead offer a short phrase as an answer.
    """

    name = "Contact Other Plane"
    level = 5
    casting_time = "1 minute"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "1 minute"
    ritual = True
    magic_school = "Divination"
    classes = ("Warlock", "Wizard")


class Contagion(Spell):
    """Your touch inflicts disease.  Make a melee spell attack against a
    creature within your reach. On a hit, you afflict the creature
    with a disease of your choice from any of the ones described
    below.

    At the end of each of the target's turns, it must make a
    Constitution saving throw. After failing three of these saving
    throws, the disease's effects last for the duration, and the
    creature stops making these saves. After succeeding on three of
    these saving throws, the creature recovers from the disease, and
    the spell ends.

    Since this spell induces a natural disease in its target, any
    effect that removes a disease or otherwise ameliorates a disease's
    effects apply to it.

    Blinding Sickness
      Pain grips the creature's mind, and its eyes turn milky
      white. The creature has disadvantage on Wisdom checks and Wisdom
      saving throws and is blinded.
    Filth Fever
      A raging fever sweeps through the creature's body. The creature
      has disadvantage on Strength checks, Strength saving throws, and
      attack rolls that use Strength.
    Flesh Rot
      The creature's flesh decays. The creature has disadvantage on
      Charisma checks and vulnerability to all damage.
    Mindfire
      The creature's mind becomes feverish. The creature has
      disadvantage on Intelligence checks and Intelligence saving
      throws, and the creature behaves as if under the effects of the
      confusion spell during combat.
    Seizure
      The creature is overcome with shaking. The creature has
      disadvantage on Dexterity checks, Dexterity saving throws, and
      attack rolls that use Dexterity.
    Slimy Doom
      The creature begins to bleed uncontrollably. The creature has
      disadvantage on Constitution checks and Constitution saving
      throws. In addition, whenever the creature takes damage, it is
      stunned until the end of its next turn.

    """

    name = "Contagion"
    level = 5
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "7 days"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric", "Druid")


class Contingency(Spell):
    """Choose a spell of 5th level or lower that you can cast, that has a
    casting time of 1 action, and that can target you.  You cast that
    spell called "the contingent spell" as part of casting
    contingency, expending spell slots for both, but the contingent
    spell doesn't come into effect. Instead, it takes effect when a
    certain circumstance occurs. You describe that circumstance when
    you cast the two spells. For example, a contingency cast with
    water breathing might stipulate that water breathing comes into
    effect when you are engulfed in water or a similar liquid.

    The contingent spell takes effect immediately after the
    circumstance is met for the first time, whether or not you want
    it to. and then contingency ends.

    The contingent spell takes effect only on you, even if it can
    normally target others. You can use only one contingency spell at
    a time.  If you cast this spell again, the effect of another
    contingency spell on you ends. Also, contingency ends on you if
    its material component is ever not on your person.

    """

    name = "Contingency"
    level = 6
    casting_time = "10 minutes"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = (
        "A statuette of yourself carved from ivory and decorated with gems worth at"
        " least 1,500 gp"
    )
    duration = "10 days"
    ritual = False
    magic_school = "Evocation"
    classes = ("Wizard",)


class ContinualFlame(Spell):
    """A flame, equivalent in brightness to a torch, springs forth from an
    object that you touch.  The effect looks like a regular flame, but
    it creates no heat and doesn't use oxygen. A continual flame can
    be covered or hidden but not smothered or quenched.

    """

    name = "Continual Flame"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Ruby dust worth 50 gp, which the spell consumes"
    duration = "Until dispelled"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric", "Wizard")


class ControlFlames(Spell):
    """You choose a nonmagical flame that you can see within range and
    that fits within a 5-foot cube. You affect it in one of the
    following ways:

    - You instantaneously expand the flame 5 feet in one direction,
      provided that wood or other fuel is present in the new location.
    - You instantaneously extinguish the flames within the cube.
    - You double or halve the area of bright light and dim light cast
      by the flame, change its color, or both. The change lasts for 1
      hour.
    - You cause simple shapes - such as the vague form of a creature,
      an inanimate object, or a location - to appear within the flames
      and animate as you like. The shapes last for 1 hour.  If you
      cast this spell multiple times, you can have up to three
      non-instantaneous

    """

    name = "Control Flames"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("S",)
    materials = ""
    duration = "Instantaneous or 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Wizard")


class ControlWater(Spell):
    """Until the spell ends, you control any freestanding water inside an
    area you choose that is a cube up to 100 feet on a side.  You can
    choose from any of the following effects when you cast this
    spell. As an action on your turn, you can repeat the same effect
    or choose a different one.

    Flood
      You cause the water level of all standing water in the area to
      rise by as much as 20 feet. If the area includes a shore, the
      flooding water spills over onto dry land. If you choose an area
      in a large body of water, you instead create a 20-foot tall wave
      that travels from one side of the area to the other and then
      crashes down. Any Huge or smaller vehicles in the wave's path
      are carried with it to the other side. Any Huge or smaller
      vehicles struck by the wave have a 25 percent chance of
      capsizing.

      The water level remains elevated until the spell ends
      or you choose a different effect. If this effect produced a
      wave, the wave repeats on the start of your next turn while the
      flood effect lasts.
    Part Water
      You cause water in the area to move apart and create a
      trench. The trench extends across the spell's area, and the
      separated water forms a wall to either side.  The trench remains
      until the spell ends or you choose a different effect. The water
      then slowly fills in the trench over the course of the next
      round until the normal water level is restored.
    Redirect Flow
      You cause flowing water in the area to move in a direction you
      choose, even if the water has to flow over obstacles, up walls,
      or in other unlikely directions. The water in the area moves as
      you direct it, but once it moves beyond the spell's area, it
      resumes its flow based on the terrain conditions. The water
      continues to move in the direction you chose until the spell
      ends or you choose a different effect.
    Whirlpool
      This effect requires a body of water at least 50 feet square and
      25 feet deep. You cause a whirlpool to form in the center of the
      area. The whirlpool forms a vortex that is 5 feet wide at the
      base, up to 50 feet wide at the top, and 25 feet tall. Any
      creature or object in the water and within 25 feet of the vortex
      is pulled 10 feet toward it. A creature can swim away from the
      vortex by making a Strength (Athletics) check against your spell
      save DC.

      When a creature enters the vortex for the first time on a turn
      or starts its turn there, it must make a Strength saving
      throw. On a failed save, the creature takes 2d8 bludgeoning
      damage and is caught in the vortex until the spell ends.  On a
      successful save, the creature takes half damage, and isn't
      caught in the vortex. A creature caught in the vortex can use
      its action to try to swim away from the vortex as described
      above, but has disadvantage on the Strength (Athletics) check to
      do so.

      The first time each turn that an object enters the
      vortex, the object takes 2d8 bludgeoning damage, this damage
      occurs each round it remains in the vortex.

    """

    name = "Control Water"
    level = 4
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ("V", "S", "M")
    materials = "A drop of water and a pinch of dust"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Cleric", "Druid", "Wizard")


class ControlWeather(Spell):
    """You take control of the weather within 5 miles of you for the
    duration.

    You must be outdoors to cast this spell. Moving to a place where
    you don't have a clear path to the sky ends the spell early.

    When you cast the spell, you change the current weather
    conditions, which are determined by the DM based on the climate
    and season. You can change precipitation, temperature, and
    wind. It takes 1d4 x 10 minutes for the new conditions to take
    effect. Once they do so, you can change the conditions again. When
    the spell ends, the weather gradually returns to normal.

    When you change the weather conditions, find a current condition
    on the following tables and change its stage by one, up or
    down. When changing the wind, you can change its direction.

    **Precipitation**
    Stage 1 – Clear,
    Stage 2 – Light clouds,
    Stage 3 – Overcast or ground fog,
    Stage 4 – Rain, hail or snow,
    Stage 5 – Torrential rain, driving hail or blizzard

    **Temperature**
    Stage 1 – Unbearable heat,
    Stage 2 – Hot,
    Stage 3 – Warm,
    Stage 4 – Cool,
    Stage 5 – Cold,
    Stage 6 – Arctic cold

    **Wind**
    Stage 1 – Calm,
    Stage 2 – Moderate wind,
    Stage 3 – Strong wind,
    Stage 4 – Gale,
    Stage 5 – Storm

    """

    name = "Control Weather"
    level = 8
    casting_time = "10 minutes"
    casting_range = "Self (5-mile radius)"
    components = ("V", "S", "M")
    materials = ""
    duration = "Concentration, up to 8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Cleric", "Druid", "Wizard")


class ControlWinds(Spell):
    """You take control of the air in a 100-foot cube that you can see
    within range.  Choose one of the following effects when you cast
    the spell. The effect lasts for the spell's duration, unless you
    use your action on a later turn to switch to a different
    effect. You can also use your action to temporarily halt the
    effect or to restart one you've halted.

    Gusts. A wind picks up within the cube, continually blowing in a
    horizontal direction that you choose. You choose the intensity of
    the wind: calm, moderate, or strong. If the wind is moderate or
    strong, ranged weapon attacks that pass through it or that are
    made against targets within the cube have disadvantage on their
    attack rolls. If the wind is strong, any creature moving against
    the wind must spend 1 extra foot of movement for each foot moved.

    Downdraft. You cause a sustained blast of strong wind to blow
    downward from the top of the cube. Ranged weapon attacks that pass
    through the cube or that are made against targets within it have
    disadvantage on their attack rolls. A creature must make a
    Strength saving throw if it flies into the cube for the first time
    on a turn or starts its turn there flying. On a failed save, the
    creature is knocked prone.

    Updraft. You cause a sustained updraft within the cube, rising
    upward from the cube's bottom edge. Creatures that end a fall
    within the cube take only half damage from the fall. When a
    creature in the cube makes a vertical jump, the creature can jump
    up to 10 feet higher than normal.

    """

    name = "Control Winds"
    level = 5
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Wizard")


class CordonOfArrows(Spell):
    """You plant four pieces of nonmagical ammunition – arrows or crossbow bolts – in
    the ground within range and lay magic upon them to protect an area.
    Until the
    spell ends, whenever a creature other than you comes within 30 feet of the
    ammunition for the first time on a turn or ends its turn there, one piece of
    ammunition flies up to strike it. The creature must succeed on a Dexterity
    saving throw or take 1d6 piercing damage. The piece of ammunition is then
    destroyed. The spell ends when no ammunition remains.

    When you cast this
    spell, you can designate any creatures you choose, and the spell ignores them.


    **At Higher Levels:** When you cast this spell using a spell slot of 3rd level or
    higher, the amount of ammunition that can be affected increases by two for each
    slot level above 2nd.
    """

    name = "Cordon Of Arrows"
    level = 2
    casting_time = "1 action"
    casting_range = "5 feet"
    components = ("V", "S", "M")
    materials = "Four or more arrows or bolts"
    duration = "8 hours"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Ranger",)


class Counterspell(Spell):
    """1 reaction, which you take when you see a creature within 60 feet of you casting
    a spell

    You attempt to interrupt a creature in the process of casting a spell.
    If the creature is casting a spell of 3rd level or lower, its spell fails and
    has no effect. If it is casting a spell of 4th level or higher, make an ability
    check using your spellcasting ability.  The DC equals 10+ the spell's level. On
    a success, the creature's spell fails and has no effect.

    **At Higher Levels:** When
    you cast this spell using a spell slot of 4th level or higher, the interrupted
    spell has no effect if its level is less than or equal to the level of the spell
    slot you used.
    """

    name = "Counterspell"
    level = 3
    casting_time = "Special"
    casting_range = "60 feet"
    components = ("S",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Sorcerer", "Warlock", "Wizard")


class CreateBonfire(Spell):
    """You create a bonfire on ground that you can see within range. Until
    the spell ends, the magic bonfire fills a 5-foot cube. Any
    creature in the bonfire's space when you cast the spell must
    succeed on a Dexterity saving throw or take 1d8 fire damage. A
    creature must also make the saving throw when it moves into the
    bonfire's space for the first time on a turn or ends its turn
    there.  The bonfire ignites flammable objects in its area that
    aren't being worn or carried.  The spell's damage increases by 1d8
    when you reach 5th level (2d8), 11th level (3d8), and 17th level
    (4d8).

    """

    name = "Create Bonfire"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class CreateFoodAndWater(Spell):
    """You create 45 pounds of food and 30 gallons of water on the ground or in
    containers within range, enough to sustain up to fifteen humanoids or five
    steeds for 24 hours. The food is bland but nourishing, and spoils if uneaten
    after 24 hours. The water is clean and doesn't go bad.
    """

    name = "Create Food And Water"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric", "Paladin")


class CreateHomunculus(Spell):
    """While speaking an intricate incantation, you cut yourself with a jewel-encrusted
    dagger, taking 2d4 piercing damage that can't be reduced in any way. You then
    drip your blood on the spell's other components and touch them, transforming
    them into a special construct called a homunculus. The statistics of the
    homunculus are in the Monster Manual. It is your faithful companion, and it dies
    if you die. Whenever you finish a long rest, you can spend up to half your Hit
    Dice if the homunculus is on the same plane of existence as you. When you do so,
    roll each die and add your Constitution modifier to it. Your hit point maximum
    is reduced by the total, and the homunculus's hit point maximum and current hit
    points are
    both increased by it. This process can reduce you to no lower than 1
    hit point. and the change to your and the homunculus's hit points ends when you
    finish your next long rest. The reduction to your hit point maximum can't be
    removed by any means before then, except by the homunculus‘s death. You can have
    only one homunculus at a time. If you cast this spell while your homunculus
    lives, the spell fails.
    """

    name = "Create Homunculus"
    level = 6
    casting_time = "1 hour"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "Clay, ash, and mandrake root, all of which the spell consumes, and a"
        " jewel-encrusted dagger worth at least 1,000 gp"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Wizard",)


class CreateOrDestroyWater(Spell):
    """You either create or destroy water.

    Create Water
      You create up to 10 gallons of clean water within range in an
      open container. Alternatively, the water falls as rain in a
      30-foot cube within range, extinguishing exposed flames in the
      area.

    Destroy Water
      You destroy up to 10 gallons of water in an open container
      within range. Alternatively, you destroy fog in a 30-foot cube
      within range.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 2nd level or higher, you create or destroy 10 additional
    gallons of water, or the size of the cube increases by 5 feet, for
    each slot level above 1st.

    """

    name = "Create Or Destroy Water"
    level = 1
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = (
        "A drop of water if creating water or a few grains of sand if destroying it"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Cleric", "Druid")


class CreateUndead(Spell):
    """You can cast this spell only at night. Choose up to three corpses of Medium or
    Small humanoids within range. Each corpse becomes a ghoul under your control.
    (The DM has game statistics for these creatures.)

    As a bonus action on each of
    your turns, you can mentally command any creature you animated with this spell
    if the creature is within 120 feet of you (if you control multiple creatures,
    you can command any or all of them at the same time, issuing the same command to
    each one). You decide what action the creature will take and where it will move
    during its next turn, or you can issue a general command, such as to guard a
    particular chamber or corridor. If you issue no commands, the creature only
    defends itself against hostile creatures. Once given an order, the creature
    continues to follow it until its task is complete.

    The creature is under your
    control for 24 hours, after which it stops obeying any command you have given
    it. To maintain control of the creature for another 24 hours, you must cast this
    spell on the creature before the current 24-hour period ends. This use of the
    spell reasserts your control over up to three creatures you have animated with
    this spell, rather than animating new ones.

    **At Higher Levels:** When you cast
    this spell using a 7th-level spell slot, you can animate or reassert control
    over four ghouls.
    When you cast this spell using an 8th-level spell slot, you
    can animate or reassert control over five ghouls or two ghasts or wights.
    When
    you cast this spell using a 9th-level spell slot, you can animate or reassert
    control over six ghouls, three ghasts or wights, or two mummies.
    """

    name = "Create Undead"
    level = 6
    casting_time = "1 minute"
    casting_range = "10 feet"
    components = ("V", "S", "M")
    materials = (
        "One clay pot filled with grave dirt, one clay pot filled with brackish water,"
        " and one 150 gp black onyx stone for each corpse"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric", "Warlock", "Wizard")


class Creation(Spell):
    """You pull wisps of shadow material from the Shadowfell to create a nonliving
    object of vegetable matter within range: soft goods, rope, wood, or something
    similar. You can also use this spell to create mineral objects such as stone,
    crystal, or metal. The object created must be no larger than a 5-foot cube, and
    the object must be of a form and material that you have seen before.

    The
    duration depends on the object's material. If the object is composed of multiple
    materials, use the shortest duration.

    Material - Duration
    Vegetable matter - 1 day
    Stone/crystal - 12 hours
    Precious metals - 1 hour
    Gems - 10 minutes

    Adamantine/Mithral - 1 minute

    Using any material created by this spell as another spell's
    material component causes that spell to fail.

    **At Higher Levels:** When you cast this spell using a spell slot of
     6th level or higher, the cube increases by 5 feet for each slot
     level above 5th.

    """

    name = "Creation"
    level = 5
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A tiny piece of matter of the same type of the item you plan to create"
    duration = "Special"
    ritual = False
    magic_school = "Illusion"
    classes = ("Sorcerer", "Wizard")


class CrownOfMadness(Spell):
    """One humanoid of your choice that you can see within range must succeed on a
    Wisdom saving throw or become charmed by you for the duration.
    While the target
    is charmed in this way, a twisted crown of jagged iron appears on its head, and
    a madness glows in its eyes.

    The charmed target must use its action before
    moving on each of its turns to make a melee attack against a creature other than
    itself that you mentally choose. The target can act normally on its turn if you
    choose no creature or if none are within its reach.

    On your subsequent turns,
    you must use your action to maintain control over the target, or the spell
    ends. Also, the target can make a Wisdom saving throw at the end of each of its
    turns. On a success, the spell ends.
    """

    name = "Crown Of Madness"
    level = 2
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class CrownOfStars(Spell):
    """Seven star-like motes of light appear and orbit your head until the
    spell ends.  You can use a bonus action to send one of the motes
    streaking toward one creature or object within 120 feet of
    you. When you do so, make a ranged spell attack. On a hit. the
    target takes 4d12 radiant damage. Whether you hit or miss, the
    mote is expended. The spell ends early if you expend the last
    mote. If you have four or more motes remaining, they shed bright
    light in a 30-foot radius and dim light for an additional 30
    feet. Ifyou have one to three motes remaining, they shed dim light
    in a 30-foot radius.

    **At Higher Levels:** When you cast this spell using a spell slot of
    8th level or higher, the number of motes created increases by two
    for each slot level above 7th.

    """

    name = "Crown Of Stars"
    level = 7
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Warlock", "Wizard")


class CrusadersMantle(Spell):
    """Holy power radiates from you in an aura with a 30-foot radius,
    awakening boldness in friendly creatures. Until the spell ends,
    the aura moves with you, centered on you. While in the aura, each
    nonhostile creature in the aura (including you) deals an extra 1d4
    radiant damage when it hits with a weapon attack.

    """

    name = "Crusaders Mantle"
    level = 3
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Paladin",)


class CureWounds(Spell):
    """A creature you touch regains a number of hit points equal to 1d8 + your
    spellcasting ability modifier. This spell has no effect on undead or constructs.


    **At Higher Levels:** When you cast this spell using a spell slot of 2nd level or
    higher, the healing increases by 1d8 for each slot level above 1st.
    """

    name = "Cure Wounds"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Ranger")



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability
class Quasit(Monster):
    """
    Shapechanger.
      The quasit can use its action to polymorph into a beast form that
      resembles a bat (speed 10 ft. fly 40 ft.), a centipede (40 ft., climb
      40 ft.), or a toad (40 ft., swim 40 ft.), or back into its true form .
      Its statistics are the same in each form, except for the speed changes
      noted. Any equipment it is wearing or carrying isn't transformed . It
      reverts to its true form if it dies.
    Magic Resistance.
      The quasit has advantage on saving throws against spells and other
      magical effects.
    Variant: Familiar.
      The quasit can serve another creature as a familiar, forming a
      telepathic bond with its willing master. While the two are bonded, the
      master can sense what the quasit senses as long as they are within 1
      mile of each other. While the quasit is within 10 feet of its master,
      the master shares the quasit's Magic Resistance trait. At any time and
      for any reason, the quasit can end its service as a familiar, ending
      the telepathic bond.
    Claw (Bite in Beast Form).
      Melee Weapon Attack: +4 to hit, reach 5 ft ., one target. Hit: 5 (1d4
      + 3) piercing damage, and the target must succeed on a DC 10
      Constitution saving throw or take 5 (2d4) poison damage and become
      poisoned for 1 minute. The target can repeat the saving throw at the
      end of each of its turns, ending the effect on itself on a success.
    Scare.
      One creature of the quasit's choice within 20 ft. of it must succeed
      on a DC 10 Wisdom saving throw or be frightened for 1 minute. The
      target can repeat the saving throw at the end of each of its turns,
      with disadvantage if the quasit is within line of sight, ending the
      effect on itself on a success.
    Invisibility.
      The quasit magically turns invisible until it attacks or uses Scare,
      or until its concentration ends (as if concentrating on a spell). Any
      equipment the quasit wears or carries is invisible with it.
    """
    name = "Quasit"
    description = "Tiny fiend, chaotic evil"
    challenge_rating = 1
    armor_class = 13
    skills = "Stealth +5"
    senses = "Darkvision 120 ft., Passive Perception 10"
    languages = "Abyssal, Common"
    strength = Ability(5)
    dexterity = Ability(17)
    constitution = Ability(10)
    intelligence = Ability(7)
    wisdom = Ability(10)
    charisma = Ability(10)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 7
    hit_dice = "3d4"
    spells = []


class Quipper(Monster):
    """
    Blood Frenzy.
      The quipper has advantage on melee attack rolls against any creature
      that doesn't have all its hit points.
    Water Breathing.
      The quipper can breathe only underwater.
    Bite.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 1
      piercing damage.
    """
    name = "Quipper"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 13
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 8"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(16)
    constitution = Ability(9)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(2)
    speed = 0
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 1
    hit_dice = "1d4"
    spells = []



__all__ = (
    "__version__",
    "Character",
    "weapons",
    "features",
    "mechanics",
    "race",
    "background",
    "spells",
    "import_homebrew",
)

from dungeonsheets import background, features, race, spells, weapons, mechanics
from dungeonsheets.character import Character
from dungeonsheets.content_registry import import_homebrew
from dungeonsheets.content import __version__



from dungeonsheets import spells
from dungeonsheets.features.features import Feature


# PHB
class ArcaneRecovery(Feature):
    """You have learned to regain some of your magical energy by studying
    your spellbook. Once per day when you finish a short rest, you can
    choose expended spell slots to recover. The spell slots can have a
    combined level that is equal to or less than half your wizard
    level (rounded up), and none of the slots can be 6th level or
    higher. For example, if you're a 4th-level wizard, you can recover
    up to two levels worth o f spell slots. You can recover either a
    2nd-level spell slot or two 1st-level spell slots

    """

    name = "Arcane Recovery"
    source = "Wizard"


class SpellMastery(Feature):
    """At 18th level, you have achieved such mastery over certain spells that you
    can cast them at will. Choose a 1st-level wizard spell and a 2nd-level
    wizard spell that are in your spellbook. You can cast those spells at their
    lowest level without expending a spell slot when you have them prepared. If
    you want to cast either spell at a higher level, you must expend a spell
    slot as normal. By spending 8 hours in study, you can exchange one or both
    o f the spells you chose for different spells of the same levels.

    """

    name = "Spell Mastery"
    source = "Wizard"


class SignatureSpells(Feature):
    """When you reach 20th level, you gain mastery over two powerful spells and
    can cast them with little effort. Choose two 3rd-level wizard spells in
    your spellbook as your signature spells. You always have these spells
    prepared, they don't count against the number of spells you have prepared,
    and you can cast each of them once at 3rd level without expending a spell
    slot. When you do so, you can't do so again until you finish a short or
    long rest. If you want to cast either spell at a higher level, you must
    expend a spell slot as normal.

    """

    name = "Signature Spells"
    source = "Wizard"


# Abjuration
class AbjurationSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an abjuration spell into your spellbook is halved.

    """

    name = "Abjuration Savant"
    source = "Wizard (School of Abjuration)"


class ArcaneWard(Feature):
    """Starting at 2nd level, you can weave magic around yourself for
    protection. When you cast an abjuration spell of 1st level or higher, you
    can simultaneously use a strand of the spell's magic to create a magical
    ward on yourself that lasts until you finish a long rest. The ward has hit
    points equal to twice your wizard level + your Intelligence
    modifier. Whenever you take damage, the ward takes the damage instead. If
    this damage reduces the ward to 0 hit points, you take any remaining
    damage. While the ward has 0 hit points, it can't absorb damage, but its
    magic remains. Whenever you cast an abjuration spell of 1st level or
    higher, the ward regains a number of hit points equal to twice the level of
    the spell. Once you create the ward, you can't create it again until you
    finish a long rest

    """

    name = "Arcane Ward"
    source = "Wizard (School of Abjuration)"


class ProjectedWard(Feature):
    """Starting at 6th level, when a creature that you can see within 30 feet of
    you takes damage, you can use your reaction to cause your Arcane Ward to
    absorb that damage. If this damage reduces the ward to 0 hit points, the
    warded creature takes any remaining damage

    """

    name = "Projected Ward"
    source = "Wizard (School of Abjuration)"


class ImprovedAbjuration(Feature):
    """Beginning at 10th level, when you cast an abjuration spell that requires
    you to make an ability check as a part of casting that spell (as in
    counterspell and dispel magic), you add your proficiency bonus to that
    ability check.

    """

    name = "Improved Abjuration"
    source = "Wizard (School of Abjuration)"


class SpellResistance(Feature):
    """Starting at 14th level, you have advantage on saving throws against
    spells. Furthermore, you have resistance against the damage of spells

    """

    name = "Spell Resistance"
    source = "Wizard (School of Abjuration)"


# Conjuration
class ConjurationSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy a conjuration spell into your spellbook is halved.

    """

    name = "Conjuration Savant"
    source = "Wizard (School of Conjuration)"


class MinorIllusion(Feature):
    """Starting at 2nd level when you select this school, you can use your action
    to conjure up an inanimate object in your hand or on the ground in an
    unoccupied space that you can see within 10 feet of you. This object can be
    no larger than 3 feet on a side and weigh no more than 10 pounds, and its
    form must be that of a nonmagical object that you have seen. The object is
    visibly magical, radiating dim light out to 5 feet. The object disappears
    after 1 hour, when you use this feature again, or if it takes any damage.

    """

    name = "Minor Illusion"
    source = "Wizard (School of Conjuration)"


class BenignTransposition(Feature):
    """Starting at 6th level, you can use your action to teleport up to 30 feet to
    an unoccupied space that you can see. Alternatively, you can choose a space
    within range that is occupied by a Small or Medium creature. If that
    creature is willing, you both teleport, swapping places. Once you use this
    feature, you can't use it again until you finish a long rest or you cast a
    conjuration spell of 1st level or higher.

    """

    name = "Benign Transposition"
    source = "Wizard (School of Conjuration)"


class FocusedConjuration(Feature):
    """Beginning at 10th level, while you are concentrating on a conjuration
    spell, your concentration can't be broken as a result of taking damage

    """

    name = "Focused Conjuration"
    source = "Wizard (School of Conjuration)"


class DurableSummons(Feature):
    """Starting at 14th level, any creature that you summon or create with a
    conjuration spell has 30 temporary hit points

    """

    name = "Durable Summons"
    source = "Wizard (School of Conjuration)"


# Divination
class DivinationSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy a divination spell into your spellbook is halved.  P

    """

    name = "Divination Savant"
    source = "Wizard (School of Divination)"


class Portent(Feature):
    """Starting at 2nd level when you choose this school, glimpses of the future
    begin to press in on your awareness. When you finish a long rest, roll two
    d20s and record the numbers rolled. You can replace any attack roll, saving
    throw, or ability check made by you or a creature that you can see with one
    of these foretelling rolls. You must choose to do so before the roll, and
    you can replace a roll in this way only once per turn. Each foretelling
    roll can be used only once. When you finish a long rest, you lose any
    unused foretelling rolls

    """

    name = "Portent"
    source = "Wizard (School of Divination)"


class ExpertDivination(Feature):
    """Beginning at 6th level, casting divination spells comes so easily to you
    that it expends only a fraction of your spellcasting efforts. When you cast
    a divination spell of 2nd level or higher using a spell slot, you regain
    one expended spell slot. The slot you regain must be of a level lower than
    the spell you cast and can't be higher than 5th level.

    """

    name = "Expert Divination"
    source = "Wizard (School of Divination)"


class TheThirdEye(Feature):
    """Starting at 10th level, you can use your action to increase your powers o f
    perception. When you do so, choose one of the following benefits, which
    lasts until you are incapacitated or you take a short or long rest. You
    can't use the feature again until you finish a rest.

    **Darkvision**: You gain darkvision out to a range of 60 feet, as described
    in chapter 8.

    **Ethereal Sight**: You can see into the Ethereal Plane within 60
    feet of you.

    **Greater Comprehension**: You can read any language.

    **See Invisibility**: You can see invisible creatures and objects within 10
    feet of you that are within line of sight

    """

    name = "The Third Eye"
    source = "Wizard (School of Divination)"


class GreaterPortent(Feature):
    """Starting at 14th level, the visions in your dreams intensify and paint a
    more accurate picture in your mind of what is to come. You roll three d20s
    for your Portent feature, rather than two

    """

    name = "Greater Portent"
    source = "Wizard (School of Divination)"


# Enchantment
class EnchantmentSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an enchantment spell into your spellbook is halved

    """

    name = "Enchantment Savant"
    source = "Wizard (School of Enchantment)"


class HypnoticGaze(Feature):
    """Starting at 2nd level when you choose this school, your soft words and
    enchanting gaze can magically enthrall another creature. As an action,
    choose one creature that you can see within 5 feet of you. If the target
    can see or hear you, it must succeed on a Wisdom saving throw against your
    wizard spell save DC or be charmed by you until the end of your next
    turn. The charmed creature's speed drops to 0, and the creature is
    incapacitated and visibly dazed.

    On subsequent turns, you can use your action to maintain this effect,
    extending its duration until the end of your next turn. However, the effect
    ends if you move more than 5 feet away from the creature, if the creature
    can neither see nor hear you, or if the creature takes damage. Once the
    effect ends, or if the creature succeeds on its initial saving throw
    against this effect, you can't use this feature on that creature again
    until you finish a long rest

    """

    name = "Hypnotic Gaze"
    source = "Wizard (School of Enchantment)"


class InstinctiveGaze(Feature):
    """Beginning at 6th level, when a creature you can see within 30 feet of you
    makes an attack roll against you, you can use your reaction to divert the
    attack, provided that another creature is within the attack's range. The
    attacker must make a W isdom saving throw against your wizard spell save
    DC. On a failed save, the attacker must target the creature that is closest
    to it, not including you or itself. If multiple creatures are closest, the
    attacker chooses which one to target. On a successful save, you can't use
    this feature on the attacker again until you finish a long rest. You must
    choose to use this feature before knowing whether the attack hits or
    misses. Creatures that can't be charmed are immune to this effect.

    """

    name = "Instinctive Gaze"
    source = "Wizard (School of Enchanment)"


class SplitEnchantment(Feature):
    """Starting at 10th level, when you cast an enchantment spell of 1st level or
    higher that targets only one creature, you can have it target a second
    creature.

    """

    name = "Split Enchantment"
    source = "Wizard (School of Enchanment)"


class AlterMemories(Feature):
    """At 14th level, you gain the ability to make a creature unaware of your
    magical influence on it. When you cast an enchantment spell to charm one or
    more creatures, you can alter one creature's understanding so that it
    remains unaware of being charmed. Additionally, once before the spell
    expires, you can use your action to try to make the chosen creature forget
    some of the time it spent charmed. The creature must succeed on an
    Intelligence saving throw against your wizard spell save DC or lose a
    number of hours of its memories equal to 1 + your Charisma modifier
    (minimum 1). You can make the creature forget less time, and the amount of
    time can't exceed the duration of your enchantment spell.

    """

    _name = "Alter Memories"
    source = "Wizard (School of Enchanment)"

    @property
    def name(self):
        num = 1 + max(0, self.owner.charisma.modifier)
        return self._name + " ({:d} hours)".format(num)


# Evocation
class EvocationSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an evocation spell into your spellbook is halved.

    """

    name = "Evocation Savant"
    source = "Wizard (School of Evocation)"


class SculptSpells(Feature):
    """Beginning at 2nd level, you can create pockets of relative safety within the
    effects of your evocation spells. When you cast an evocation spell that
    affects other creatures that you can see, you can choose a number of them
    equal to 1 + the spell's level. The chosen creatures automatically succeed
    on their saving throws against the spell, and they take no damage if they
    would normally take half damage on a successful save

    """

    name = "Sculpt Spells"
    source = "Wizard (School of Evocation)"


class PotentCantrip(Feature):
    """Starting at 6th level, your damaging cantrips affect even creatures that
    avoid the brunt of the effect. When a creature succeeds on a saving throw
    against your cantrip, the creature takes half the cantrip's damage (if any)
    but suffers no additional effect from the cantrip

    """

    name = "Potent Cantrip"
    source = "Wizard (School of Evocation)"


class EmpoweredEvocation(Feature):
    """Beginning at 10th level, you can add your Intelligence modifier to the
    damage roll of any wizard evocation spell you cast

    """

    name = "Empowered Evocation"
    source = "Wizard (School of Evocation)"


class Overchannel(Feature):
    """Starting at 14th level, you can increase the power of your simpler
    spells. When you cast a wizard spell of 5th level or lower that deals
    damage, you can deal maximum damage with that spell. The first time you do
    so, you suffer no adverse effect.  If you use this feature again before you
    finish a long rest, you take 2d12 necrotic damage for each level of the
    spell, immediately after you cast it. Each time you use this feature again
    before finishing a long rest, the necrotic damage per spell level increases
    by 1d12. This damage ignores resistance and immunity.

    """

    name = "Overchannel"
    source = "Wizard (School of Evocation)"


# Illusion
class IllusionSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an illusion spell into your spellbook is halved.

    """

    name = "Illusion Savant"
    source = "Wizard (School of Illusion)"


class ImprovedMinorIllusion(Feature):
    """When you choose this school at 2nd level, you learn the minor illusion
    cantrip. If you already know this cantrip, you learn a different wizard
    cantrip of your choice. The cantrip doesn't count against your number of
    cantrips known. When you cast minor illusion, you can create both a sound
    and an image with a single casting o f the spell.

    """

    name = "Improved Minor Illusion"
    source = "Wizard (School of Illusion)"


class MalleableIllusions(Feature):
    """Starting at 6th level, when you cast an illusion spell that has a duration
    of 1 minute or longer, you can use your action to change the nature of that
    illusion (using the spell's normal parameters for the illusion), provided
    that you can see the illusion

    """

    name = "Malleable Illusions"
    source = "Wizard (School of Illusion)"


class IllusorySelf(Feature):
    """Beginning at 10th level, you can create an illusory duplicate of yourself
    as an instant, almost instinctual reaction to danger. When a creature makes
    an attack roll against you, you can use your reaction to interpose the
    illusory duplicate between the attacker and yourself. The attack
    automatically m isses you, then the illusion dissipates. Once you use this
    feature, you can't use it again until you finish a short or long rest.

    """

    name = "Illusory Self"
    source = "Wizard (School of Illusion)"


class IllusoryReality(Feature):
    """By 14th level, you have learned the secret of weaving shadow magic into
    your illusions to give them a semi- reality. When you cast an illusion
    spell of 1st level or higher, you can choose one inanimate, nonmagical
    object that is part of the illusion and make that object real. You can do
    this on your turn as a bonus action while the spell is ongoing. The object
    remains real for 1 minute. For example, you can create an illusion of a
    bridge over a chasm and then make it real long enough for your allies to
    cross. The object can't deal damage or otherwise directly harm anyone

    """

    name = "Illusory Reality"
    source = "Wizard (School of Illusion)"


# Necromancy
class NecromancySavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an Necromancy spell into your spellbook is halved.

    """

    name = "Necromancy Savant"
    source = "Wizard (School of Necromancy)"


class GrimHarvest(Feature):
    """At 2nd level, you gain the ability to reap life energy from creatures you
    kill with your spells. Once per turn when you kill one or more creatures
    with a spell of 1st level or higher, you regain hit points equal to twice
    the spell's level, or three times its level if the spell belongs to the
    School of Necromancy. You don't gain this benefit for killing constructs or
    undead"""

    name = "Grim Harvest"
    source = "Wizard (School of Necromancy)"


class UndeadThralls(Feature):
    """At 6th level, you add the animate dead spell to your spellbook if
    it is not there already. When you cast animate dead, you can
    target one additional corpse or pile of bones, creating another
    zombie or skeleton, as appropriate. Whenever you create an undead
    using a necromancy spell, it has additional benefits:

    - The creature's hit point maximum is increased by an amount equal
      to your wizard level.
    - The creature adds your proficiency bonus to its weapon damage
      rolls

    """

    name = "Undead Thralls"
    source = "Wizard (School of Necromancy)"
    spells_known = (spells.AnimateDead,)


class InuredToUndeath(Feature):
    """Beginning at 10th level, you have resistance to necrotic damage, and your
    hit point maximum can't be reduced. You have spent so much time dealing
    with undead and the forces that animate them that you have become inured to
    some of their worst effects

    """

    name = "Inured to Undeath"
    source = "Wizard (School of Necromancy)"


class CommandUndead(Feature):
    """Starting at 14th level, you can use magic to bring undead under your
    control, even those created by other wizards. As an action, you can choose
    one undead that you can see within 60 feet of you. That creature must make
    a Charisma saving throw against your wizard spell save DC. If it succeeds,
    you can't use this feature on it again. If it fails, it becomes friendly to
    you and obeys your commands until you use this feature again. Intelligent
    undead are harder to control in this way.  If the target has an
    Intelligence of 8 or higher, it has advantage on the saving throw. If it
    fails the saving throw and has an Intelligence of 12 or higher, it can
    repeat the saving throw at the end of every hour until it succeeds and
    breaks free

    """

    name = "Command Undead"
    source = "Wizard (School of Necromancy)"


# Transmutation
class TransmutationSavant(Feature):
    """Beginning when you select this school at 2nd level, the gold and time you
    must spend to copy an Transmutation spell into your spellbook is halved.

    """

    name = "Transmutation Savant"
    source = "Wizard (School of Transmutation)"


class MinorAlchemy(Feature):
    """Starting at 2nd level when you select this school, you can temporarily alter
    the physical properties of one nonmagical object, changing it from one
    substance into another. You perform a special alchemical procedure on one
    object composed entirely of wood, stone (but not a gemstone), iron, copper,
    or silver, transforming it into a different one of those materials. For
    each 10 minutes you spend performing the procedure, you can transform up to
    1 cubic foot of material. After 1 hour, or until you lose your
    concentration (as if you were concentrating on a spell), the material
    reverts to its original substance

    """

    name = "Minor Alchemy"
    source = "Wizard (School of Transmutation)"


class TransmutersStone(Feature):
    """Starting at 6th level, you can spend 8 hours creating a
    transmuter's stone that stores transmutation magic. You can
    benefit from the stone yourself or give it to another creature. A
    creature gains a benefit of your choice as long as the stone is in
    the creature's possession. When you create the stone, choose the
    benefit from the following options:

    - Darkvision out to a range of 60 feet, as described in chapter 8
    - An increase to speed of 10 feet while the creature is
      unencumbered
    - Proficiency in Constitution saving throws
    - Resistance to acid, cold, fire, lightning, or thunder damage
      (your choice whenever you choose this benefit)

    Each time you cast a transmutation spell of 1st level or higher,
    you can change the effect of your stone if the stone is on your
    person. If you create a new transmuter's stone, the previous one
    ceases to function

    """

    name = "Transmuter's Stone"
    source = "Wizard (School of Transmutation)"


class Shapechanger(Feature):
    """At 10th level, you add the polymorph spell to your spellbook, if it
    is not there already. You can cast polymorph without expending a
    spell slot. When you do so, you can target only yourself and
    transform into a beast whose challenge rating is 1 or lower. Once
    you cast polymorph in this way, you can't do so again until you
    finish a short or long rest, though you can still cast it normally
    using an available spell slot

    """

    name = "Shapechanger"
    source = "Wizard (School of Transmutation)"
    spells_known = spells_prepared = (spells.Polymorph,)


class MasterTransmuter(Feature):
    """Starting at 14th level, you can use your action to consume the reserve of
    transmutation magic stored within your transmuter's stone in a single
    burst. When you do so, choose one of the following effects. Your
    transmuter's stone is destroyed and can't be remade until you finish a long
    rest.

    **Major Transformation**: You can transmute one nonmagical object-no
    larger than a 5-foot cube-into another nonmagical object of similar size
    and mass and of equal or lesser value. You must spend 10 minutes handling
    the object to transform it.

    **Panacea**: You remove all curses, diseases, and poisons affecting a creature
    that you touch with the transmuter's stone. The creature also regains all
    its hit points.

    **Restore Life**: You cast the raise dead spell on a creature you touch
    with the transmuter's stone, without expending a spell slot or needing to
    have the spell in your spellbook.

    **Restore Youth**: You touch the transmuter's stone to a willing creature,
    and that creature's apparent age is reduced by 3d10 years, to a minimum of
    13 years. This effect doesn't extend the creature's lifespan

    """

    name = "Master Transmuter"
    source = "Wizard (School of Transmutation)"


# Bladesinging
class Bladesong(Feature):
    """Starting at 2nd level, you can invoke a secret Elven magic called
    the Bladesong, provided that you aren't wearing medium or heavy
    armor or using a shield. It graces you with supernatural speed,
    agility, and focus. You can use a bonus action to start the
    Bladesong, which lasts for 1 minute. It ends early if you are
    incapac- itated, if you don medium or heavy armor or a shield, or
    if you use two hands to make an attack with a weapon. You can also
    dismiss the Bladesong at any time you choose (no action required).

    While your Bladesong is active, you gain the following benefits:

    - You gain a bonus to your AC equal to your Intelligence modifier
      (minimum of +1).
    - Your walking speed increases by 10 feet.
    - You have advantage on Dexterity (Acrobatics) checks.
    - You gain a bonus to any Constitution saving throw you make to
      maintain your concentration on a spell. The bonus equals your
      Intelligence modifier (minimum of +l).

    You can use this feature twice. You regain all expended uses of it
    when you finish a short or long rest.

    """

    name = "Bladesong (2x/SR)"
    source = "Wizard (School of Bladesinging)"


class ExtraAttackBladesinging(Feature):
    """Starting at 6th level, you can attack twice, instead of once, whenever you
    take the Attack action on your turn.

    """

    name = "Extra Attack (2x)"
    source = "Wizard (School of Bladesinging)"


class SongOfDefense(Feature):
    """Beginning at 10th level, you can direct your magic to ab- sorb damage
    while your Bladesong is active. When you take damage, you can use your
    reaction to expend one spell slot and reduce that damage to you by an
    amount equal to five times the spell slot's level.

    """

    name = "Song of Defense"
    source = "Wizard (School of Bladesinging)"


class SongOfVictory(Feature):
    """Starting at 14th level, you add your Intelligence modifier (minimum of +1)
    to the damage of your melee weapon attacks while you r Bladesong is active

    """

    name = "Song of Victory"
    source = "Wizard (School of Bladesinging)"


# War Magic
class ArcaneDeflection(Feature):
    """At 2nd level, you have learned to weave your magic to fortify yourself
    against harm. When you are hit by an at- tack or you fail a saving throw,
    you can use your reaction to gain a +2 bonus to your AC against that attack
    or a +4 bonus to that saving throw. When you use this feature, you can't
    cast spells other than cantrips until the end of your next turn.

    """

    name = "Arcane Deflection"
    source = "Wizard (School of War Magic)"


class TacticalWit(Feature):
    """Starting at 2nd level, your keen ability to assess tactical situations
    allows you to act quickly in battle. You can give yourself a bonus to your
    initiative rolls equal to your Intelligence modifier (included in stats on
    character sheet).

    """

    name = "Tactical Wit"
    source = "Wizard (School of War Magic)"


class PowerSurge(Feature):
    """Starting at 6th level, you can store magical energy within yourself to
    later empower your damaging spells. In its stored form, this energy is
    called a power surge. You can store a maximum number of power surges equal
    to your Intelligence modifier (minimum of one). Whenever you finish a long
    rest, your number of power surges reset-s to one. Whenever you successfully
    end a spell with dispel magic or counterspel], you gain one power surge, as
    you steal magic from the spell you foiled. If you end a short rest with no
    power surges, you gain one power surge

    Once per turn when you deal damage to a creature or object with a wizard
    spell, you can spend one power surge to deal extra force damage to that
    target. The ex- tra damage equals half your wizard level.

    """

    name = "Power Surge"
    source = "Wizard (School of War Magic)"


class DurableMagic(Feature):
    """Beginning at 10th level, the magic you channel helps ward off harm. While
    you maintain concentration on a spell, you have a +2 bonus to AC and all
    saving throws.

    """

    name = "Durable Magic"
    source = "Wizard (School of War Magic)"


class DeflectingShroud(Feature):
    """At 14th level, your Arcane Deflection becomes infused with deadly
    magic. When you use your Arcane Deflec- tion feature, you can cause magical
    energy to are from you. Up to three creatures ofyour choice that you can
    see within 60 feet of you each take force damage equal to half your wizard
    level.

    """

    name = "Deflecting Shroud"
    source = "Wizard (School of War Magic)"



from dungeonsheets.spells.spells import Spell


class LegendLore(Spell):
    """Name or describe a person, place, or object. The spell brings to your mind a
    brief summary of the significant lore about the thing you named. The lore might
    consist of current tales, forgotten stories, or even secret lore that has never
    been widely known. If the thing you named isn't of legendary importance, you
    gain no information. The more information you already have about the thing, the
    more precise and detailed the information you receive is.

    The information you
    learn is accurate but might be couched in figurative language. For example, if
    you have a mysterious magic axe on hand, the spell might yield this information:
    Woe to the evildoer whose hand touches the axe, for even the haft slices the
    hand of the evil ones. Only a true Child of Stone, lover and beloved of Moradin,
    may awaken the true powers of the axe, and only with the sacred word Rudnogg on
    the lips.
    """

    name = "Legend Lore"
    level = 5
    casting_time = "10 minutes"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = (
        "Incense worth at least 250 gp, which the spell consumes, and four ivory strips"
        " worth at least 50 gp each"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Divination"
    classes = ("Bard", "Cleric", "Wizard")


class LeomundsSecretChest(Spell):
    """You hide a chest, and all its contents, on the Ethereal Plane.
    You must touch
    the chest and the miniature replica that serves as a material component for the
    spell. The chest can contain up to 12 cubic feet of nonliving material (3 feet
    by 2 feet by 2 feet).

    While the chest remains on the Ethereal Plane, you can
    use an action and touch the replica to recall the chest. It appears in an
    unoccupied space on the ground within 5 feet of you. You can send the chest back
    to the Ethereal Plane by using an action and touching both the chest and the
    replica.

    After 60 days, there is a cumulative 5 percent chance per day that the
    spell's effect ends. This effect ends if you cast this spell again, if the
    smaller replica chest is destroyed, or if you choose to end the spell as an
    action. If the spell ends and the larger chest is on the Ethereal Plane, it is
    irretrievably lost.
    """

    name = "Leomunds Secret Chest"
    level = 4
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "An exquisite chest, 3 feet by 2 feet by 2 feet, constructed from rare"
        " materials worth at least 5,000 gp, and a tiny replica made from the same"
        " materials worth at least 50 gp"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Wizard",)


class LeomundsTinyHut(Spell):
    """A 10-foot-radius immobile dome of force springs into existence around and above
    you and remains stationary for the duration. The spell ends if you leave its
    area.

    Nine creatures of Medium size or smaller can fit inside the dome with
    you. The spell fails if its area includes a larger creature or more than nine
    creatures. Creatures and objects within the dome when you cast this spell can
    move through it freely. All other creatures and objects are barred from passing
    through it. Spells and other magical effects can't extend through the dome or be
    cast through it. The atmosphere inside the space is comfortable and dry,
    regardless of the weather outside.

    Until the spell ends, you can command the
    interior to become dimly lit or dark. The dome is opaque from the outside, of
    any color you choose, but it is transparent from the inside.
    """

    name = "Leomunds Tiny Hut"
    level = 3
    casting_time = "1 minute"
    casting_range = "Self (10-foot-radius hemisphere)"
    components = ("V", "S", "M")
    materials = "A small crystal bead"
    duration = "8 hours"
    ritual = True
    magic_school = "Evocation"
    classes = ("Bard", "Wizard")


class LesserRestoration(Spell):
    """You touch a creature and can end either one disease or one condition afflicting
    it. The condition can be blinded, deafened, paralyzed, or poisoned.
    """

    name = "Lesser Restoration"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Ranger")


class Levitate(Spell):
    """One creature or object of your choice that you can see within range rises
    vertically, up to 20 feet, and remains suspended there for the duration. The
    spell can levitate a target that weighs up to 500 pounds. An unwilling creature
    that succeeds on a Constitution saving throw is unaffected.

    The target can move
    only by pushing or pulling against a fixed object or surface within reach (such
    as a wall or a ceiling), which allows it to move as if it were climbing. You
    can change the target's altitude by up to 20 feet in either direction on your
    turn. If you are the target, you can move up or down as part of your move.
    Otherwise, you can use your action to move the target, which must remain within
    the spell's range.

    When the spell ends, the target floats gently to the ground
    if it is still aloft.
    """

    name = "Levitate"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = (
        "Either a small leather loop or a piece of golden wire bent into a cup shape"
        " with a long shank on one end"
    )
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Sorcerer", "Wizard")


class LifeTransference(Spell):
    """You sacrifice some of your health to mend another creature's injuries. You take
    4d8 necrotic damage, and one creature of your choice that you can see within
    range regains a number of hit points equal to twice the necrotic damage you
    take.

    **At Higher Levels:** When you cast this spell using a spell slot of 4th
    level or higher, the damage increases by 1d8 for each slot level above 3rd.
    """

    name = "Life Transference"
    level = 3
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric", "Wizard")


class Light(Spell):
    """You touch one object that is no larger than 10 feet in any dimension. Until the
    spell ends, the object sheds bright light in a 20-foot radius and dim light for
    an additional 20 feet. The light can be colored as you like. Completely covering
    the object with something opaque blocks the light. The spell ends if you cast
    it again or dismiss it as an action.

    If you target an object held or worn by a
    hostile creature, that creature must succeed on a Dexterity saving throw to
    avoid the spell.
    """

    name = "Light"
    level = 0
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "M")
    materials = "A firefly or phosphorescent moss"
    duration = "1 hour"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard", "Cleric", "Sorcerer", "Wizard")


class LightningArrow(Spell):
    """The next time you make a ranged weapon attack during the spell's duration, the
    weapon's ammunition, or the weapon itself if it's a thrown weapon, transforms
    into a bolt of lightning. Make the attack roll as normal, The target takes 4d8
    lightning damage on a hit, or half as much damage on a miss, instead of the
    weapon's normal damage.

    Whether you hit or miss, each creature within 10 feet
    of the target must make a Dexterity saving throw. Each of these creatures takes
    2d8 lightning damage on a failed save, or half as much damage on a successful
    one.

    The piece of ammunition or weapon then returns to its normal form.

    At
    Higher Levels: When you cast this spell using a spell slot of 4th level or
    higher, the damage for both effects of the spell increases by 1d8 for each slot
    level above 3rd.
    """

    name = "Lightning Arrow"
    level = 3
    casting_time = "1 bonus action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Ranger",)


class LightningBolt(Spell):
    """A stroke of lightning forming a line of 100 feet long and 5 feet wide blasts out
    from you in a direction you choose. Each creature in the line must make a
    Dexterity saving throw. A creature takes 8d6 lightning damage on a failed save,
    or half as much damage on a successful one.

    The lightning ignites flammable
    objects in the area that aren't being worn or carried.

    **At Higher Levels:** When
    you cast this spell using a spell slot of 4th level or higher, the damage
    increases by 1d6 for each slot level above 3rd.
    """

    name = "Lightning Bolt"
    level = 3
    casting_time = "1 action"
    casting_range = "Self (100-foot line)"
    components = ("V", "S", "M")
    materials = "A bit of fur and a rod of amber, crystal, or glass"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class LightningLure(Spell):
    """You create a lash of lightning energy that strikes at one creature of your
    choice that you can see within range.
    The target must succeed on a Strength
    saving throw or be pulled up to 10 feet in a straight line toward you and then
    take 1d8 lightning damage if it is within 5 feet of you.

    **At Higher Levels:** This
    spell's damage increases by 1d8 when you reach 5th level (2d8), 11th level
    (3d8), and 17th level (4d8).
    """

    name = "Lightning Lure"
    level = 0
    casting_time = "1 action"
    casting_range = "15 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Warlock", "Wizard")


class LocateAnimalsOrPlants(Spell):
    """Describe or name a specific kind of beast or plant. Concentrating on the voice
    of nature in your surroundings, you learn the direction and distance to the
    closest creature or plant of that kind within 5 miles, if any are present.
    """

    name = "Locate Animals Or Plants"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A bit of fur from a bloodhound"
    duration = "Instantaneous"
    ritual = True
    magic_school = "Divination"
    classes = ("Bard", "Druid", "Ranger")


class LocateCreature(Spell):
    """Describe or name a creature that is familiar to you. You sense the direction to
    the creature's location, as long as that creature is within 1,000 feet of you.
    If the creature is moving, you know the direction of its movement.

    The spell
    can locate a specific creature known to you, or the nearest creature of a
    specific kind (such as a human or a unicorn), so long as you have seen such a
    creature up close – within 30 feet – at least once. If the creature you
    described or named is in a different form, such as being under the effects of a
    polymorph spell, this spell doesn't locate the creature.

    This spell can't
    locate a creature if running water at least 10 feet wide blocks a direct path
    between you and the creature.
    """

    name = "Locate Creature"
    level = 4
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A bit of fur from a bloodhound"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Divination"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Ranger", "Wizard")


class LocateObject(Spell):
    """Describe or name an object that is familiar to you. You sense the direction to
    the object's location, as long as that object is within 1,000 feet of you. If
    the object is in motion, you know the direction of its movement.

    The spell can
    locate a specific object known to you, as long as you have seen it up close –
    within 30 feet – at least once. Alternatively, the spell can locate the nearest
    object of a particular kind, such as a certain kind of apparel, jewelry,
    furniture, tool, or weapon.

    This spell can't locate an object if any thickness
    of lead, even a thin sheet, blocks a direct path between you and the object.
    """

    name = "Locate Object"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "A forked twig"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Divination"
    classes = ("Bard", "Cleric", "Druid", "Paladin", "Ranger", "Wizard")


class Longstrider(Spell):
    """You touch a creature. The target's speed increases by 10 feet until the spell
    ends.

    **At Higher Levels:** When you cast this spell using a spell slot of 2nd
    level or higher, you can target one additional creature for each slot level
    above 1st.
    """

    name = "Longstrider"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A pinch of dirt"
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Druid", "Ranger", "Wizard")
